<?php
// UNB Components
// Copyright 2007-2010 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbMessage class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbMessageRevision', false)) require_once dirname(__FILE__) . '/messagerevision.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbUser', false)) require_once dirname(__FILE__) . '/user.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';

// Enumeration of moderation flags for messages.
//
class UnbMessageModerationFlag
{
	// This message is locked and cannot be read by non-moderators (which will see an info message instead).
	const Locked = 1;
	// This message is hidden and cannot even be seen by non-moderators.
	const Hidden = 2;
	// New replies to this message (also recursive, but not across a discussion entry point)
	// need to be approved by a moderator, no matter what other rights the author would normally have.
	const EnforceApproval = 4;
}

// The message class represents a single message in the UNB system and provides methods to find
// and manipulate message instances.
//
// It depends on the UnbDatabase class for a database connection. It also uses the class
// UnbMessageRevision to manage message revisions.
//
class UnbMessage extends UnbEntity
{
	// Message ID
	private $messageId = 0;
	// Time when this message was created
	private $createdTime = 0;
	// User to whom this message is assigned, usually the user who created the message
	private $ownerUserId = 0;
	// Message ID on which this message primarily references, used for tree visualisation or replies
	private $primaryReference = 0;
	// Is this message a discussion entry point?
	private $entryPoint = false;
	// Message's page name for Wiki-like usage
	private $pageName = '';
	// Moderation flags of the entire message (see UnbMessage::MF_* constants)
	private $moderationFlags = 0;
	// Time when a user last began editing this message, used for a warning message to potential concurrent editors
	private $lastBeginEdit = 0;
	// Content of this message that shall be regarded for searching, usually a simplified plain-text version of the latest approved message revision
	private $searchOptimisedContent = '';
	// Subject of this message that shall be regarded for searching, usually of the latest approved message revision
	private $searchOptimisedSubject = '';
	// Revision that shall be regarded for searching, usually of the latest approved message revision
	private $searchRevision = 0;

	// Cached owner UnbUser object
	private $ownerUser = null;

	// Cached UnbDatabase instance
	private $db;

	// Cached read access check results
	private static $readAccessCache = array();
	// Cached alter access check results
	private static $alterAccessCache = array();
	// Cached reply access check results
	private static $replyAccessCache = array();

	private function __construct()
	{
		$this->db = UnbDatabase::GetInstance();
	}

	// Creates a new message instance.
	//
	// The newly created message instance is also stored in the database.
	//
	// subject = (string) Subject of the message's first revision or {{null}} (max. 255 characters)
	// content = (string) Content of the message's first revision (max. 16,777,215 characters)
	// data = (string) Attached file data contents or {{null}}
	// dataFileName = (string) Local file name for attached file data or {{null}} (do not use together with {{data}})
	// writeToFile = (bool) Write the attached file data into a file instead of the database or {{null}} for automatic decision
	// entryPoint = (bool) Specifies whether the new message shall be a discussion entry point.
	// primaryRef = (UnbMessage) Primary message reference for the new message
	//              (int) Message ID of the primary reference for the new message
	//
	// Throws {{UnbSecurityException}} If not logged in.
	// Throws {{UnbDatabaseException}}
	// Throws {{Exception}} If the input file could not be read or copied.
	//
	// Exceptions from UnbMessageRevision::Create() are passed through.
	//
	// Returns: The newly created UnbMessage object
	//
	// Privileges: Must be logged in
	//
	public static function Create($subject, $content, $data = null, $dataFileName = null, $writeToFile = null, $entryPoint = false, $primaryRef = null)
	{
		$primaryRefId = 0;
		if (isset($primaryRef))
		{
			$primaryRefId = $primaryRef;
			if ($primaryRefId instanceof UnbMessage)
				$primaryRefId = $primaryRefId->GetMessageId();
			if (!is_int($primaryRefId))
				throw new InvalidArgumentException('Invalid argument type: $primaryRef');
			if ($primaryRefId <= 0)
				throw new InvalidArgumentException('Argument out of range: $primaryRef');
		}

		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You must be logged in to create a message.');
		$entryPoint = $entryPoint ? true : false;

		$now = time();

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction(true);
		try
		{
			$messageId = $db->QueryValue('SELECT MAX("MessageId") FROM "#message" FOR UPDATE');
			if (!isset($messageId))
				$messageId = 0;
			$messageId++;

			$sqlCols = '"MessageId", "CreatedTime", "Owner", "EntryPoint"';
			$sqlValues = '?, ?, ?, ?';
			$sqlParams = array($messageId, UnbDatabase::ConvertDateTime($now), UnbSession::GetSessionUserId(), $entryPoint ? 1 : 0);

			if ($primaryRefId > 0)
			{
				$sqlCols .= ', "PrimaryReference"';
				$sqlValues .= ', ?';
				$sqlParams[] = $primaryRefId;
			}

			$db->Execute(
				'INSERT INTO "#message" (' . $sqlCols . ') VALUES (' . $sqlValues . ')',
				$sqlParams);

			$rev = UnbMessageRevision::Create(
				$messageId,
				$subject,
				$content,
				null,
				$data,
				$dataFileName,
				$writeToFile);

			$db->Execute(
				'UPDATE "#message" SET "SearchRevision" = ? WHERE "MessageId" = ?',
				array($rev->GetRevisionNumber(), $messageId));

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		$message = new self();
		$message->messageId = $messageId;
		$message->createdTime = $now;
		$message->ownerUserId = UnbSession::GetSessionUserId();
		$message->ownerUser = UnbSession::GetSessionUser();
		$message->entryPoint = $entryPoint;
		$message->searchRevision = $rev->GetRevisionNumber();
		return $message;
	}

	// Creates a UnbMessage object from a database record array.
	//
	// rec = (array) Database record
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	//
	// Returns: (UnbMessage) object constructed from the database record
	//
	protected static function FromRecord($rec)
	{
		if (!is_array($rec))
			throw new InvalidArgumentException('Invalid argument type: $rec');
		if (!count($rec))
			throw new InvalidArgumentException('Record array is empty');

		$message = new self();
		$message->messageId = intval($rec['MessageId']);
		$message->createdTime = UnbDatabase::ConvertDateTime($rec['CreatedTime']);
		$message->ownerUserId = intval($rec['Owner']);
		$message->primaryReference = intval($rec['PrimaryReference']);
		$message->entryPoint = intval($rec['EntryPoint']) ? true : false;
		$message->pageName = $rec['PageName'];
		$message->moderationFlags = intval($rec['ModerationFlags']);
		$message->lastBeginEdit = UnbDatabase::ConvertDateTime($rec['LastBeginEdit']);
		$message->searchOptimisedContent = $rec['SearchOptimisedContent'];
		$message->searchOptimisedSubject = $rec['SearchOptimisedSubject'];
		$message->searchRevision = intval($rec['SearchRevision']);
		return $message;
	}

	// Finds a message instance by its message ID.
	//
	// messageId = (int) Message ID
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessage object, if successful, {{null}} otherwise.
	//
	public static function FindByMessageId($messageId)
	{
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord(
			'SELECT "MessageId", "CreatedTime", "Owner", "PrimaryReference", "EntryPoint", "PageName", "ModerationFlags", "LastBeginEdit", "SearchRevision"
			FROM "#message"
			WHERE "MessageId" = ?',
			$messageId);
		// TODO: Read other Search* columns
		if (!$rec) return null;

		if (!UnbSession::IsAdministrator() &&
			UnbSession::GetSessionUserId() != intval($rec['Owner']) &&
			!self::IsReadAllowedById($messageId))
		{
			// You need read access, or be administrator or message owner instead
			return null;
		}
		if ((intval($rec['ModerationFlags']) & UnbMessageModerationFlag::Hidden) &&
			UnbSession::GetSessionUserId() != intval($rec['Owner']) &&
			!UnbSession::IsModerator() &&
			!UnbSession::IsAdministrator())
		{
			// Hidden messages can only be accessed by the message owner, moderators and administrators
			return null;
		}

		return self::FromRecord($rec);
	}

	// Finds a message instance by its page name.
	//
	// pageName = (string) Page name (max. 255 characters)
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessage object, if successful, {{null}} otherwise.
	//
	public static function FindByPageName($pageName)
	{
		if (!is_string($pageName))
			throw new InvalidArgumentException('Invalid argument type: $pageName');
		$pageName = trim($pageName);
		if (!strlen($pageName))
			throw new InvalidArgumentException('Page name is empty');
		if (mb_strlen($pageName) > 255)
			throw new LengthException('[#2] Page name is too long');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord(
			'SELECT "MessageId", "CreatedTime", "Owner", "PrimaryReference", "EntryPoint", "PageName", "ModerationFlags", "LastBeginEdit", "SearchRevision"
			FROM "#message"
			WHERE "PageName" = ?',
			$pageName);
		// TODO: Read other Search* columns
		if (!$rec) return null;

		if (!UnbSession::IsAdministrator() &&
			UnbSession::GetSessionUserId() != intval($rec['Owner']) &&
			!self::IsReadAllowedById(intval($rec['MessageId'])))
		{
			// You need read access, or be administrator or message owner instead
			return null;
		}
		if ((intval($rec['ModerationFlags']) & UnbMessageModerationFlag::Hidden) &&
			UnbSession::GetSessionUserId() != intval($rec['Owner']) &&
			!UnbSession::IsModerator() &&
			!UnbSession::IsAdministrator())
		{
			// Hidden messages can only be accessed by the message owner, moderators and administrators
			return null;
		}

		return self::FromRecord($rec);
	}

	// Gets the message ID.
	//
	// Returns: (int) Message ID
	//
	public function GetMessageId()
	{
		return $this->messageId;
	}

	// Gets the time when the message was created.
	//
	// Returns (int) UNIX timestamp
	//
	public function GetCreatedTime()
	{
		return $this->createdTime;
	}

	// Gets the user ID of the message's assigned user (owner).
	//
	// Returns: (int) Owner user ID
	//
	public function GetOwnerUserId()
	{
		return $this->ownerUserId;
	}

	// Gets the UnbUser object of the message's assigned user (owner).
	//
	// Returns: (UnbUser) Owner user
	//
	public function GetOwnerUser()
	{
		if (isset($this->ownerUser))
			return $this->ownerUser;
		$this->ownerUser = UnbUser::FindByUserId($this->ownerUserId);
		return $this->ownerUser;
	}

	// Assigns the message to another user.
	//
	// owner = (int/UnbUser) New message owner
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with administrator privilege
	//
	public function SetOwner($owner)
	{
		$ownerId = $owner;
		if ($ownerId instanceof UnbUser)
			$ownerId = $ownerId->GetUserId();
		if (!is_int($ownerId))
			throw new InvalidArgumentException('Invalid argument type: $owner');
		if ($ownerId <= 0)
			throw new InvalidArgumentException('Argument out of range: $owner');

		if (!UnbSession::IsAdministrator())
			throw new UnbSecurityException('You need the administrator privilege to set the message owner.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "Owner" = ? WHERE "MessageId" = ?',
				array($ownerId, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->ownerUserId = $ownerId;
		$this->ownerUser = $owner instanceof UnbUser ? $owner : null;
		self::ClearAccessCache();
	}

	// Gets the message ID of the message's primary reference.
	//
	// Returns: (int) Reference message ID
	//
	public function GetPrimaryReference()
	{
		return $this->primaryReference;
	}

	// Gets a value indicating whether this message is a discussion entry point.
	//
	// Returns: (bool) Message is an entry point
	//
	public function IsEntryPoint()
	{
		return $this->entryPoint;
	}

	// Specifies whether this message is a discussion entry point.
	//
	// entryPoint = (bool) New discussion entry point setting
	//
	// Throws {{UnbSecurityException}} If not logged in or no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetEntryPoint($entryPoint)
	{
		$entryPoint = $entryPoint ? true : false;

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set a message entry point.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "EntryPoint" = ? WHERE "MessageId" = ?',
				array($entryPoint ? 1 : 0, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->entryPoint = $entryPoint;
	}

	// Gets the message's page name.
	//
	// Returns: (string) Page name
	//
	public function GetPageName()
	{
		return $this->pageName;
	}

	// Sets the message's page name.
	//
	// pageName = (string) New page name or {{null}} (max. 255 characters)
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{UnbSecurityException}} If not logged in, no moderator or duplicate page name.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetPageName($pageName)
	{
		if (isset($pageName))
		{
			if (!is_string($pageName))
				throw new InvalidArgumentException('Invalid argument type: $pageName');
			$pageName = trim($pageName);
			if (mb_strlen($pageName) > 255)
				throw new LengthException('[#2] Page name is too long');
			if (!strlen($pageName))
				$pageName = null;
		}

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set a page name.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "PageName" = ? WHERE "MessageId" = ?',
				array($pageName, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			if ($this->db->IsConstraintError())
				throw new UnbSecurityException('Page name is already assigned.');
			throw $ex;
		}
		$this->pageName = $pageName;
	}

	// Gets a list of all keys that allow reading this message.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (array(int)) Keys array
	//
	public function GetReadAccessKeys()
	{
		$keys = $this->db->QueryArray(
			'SELECT k."UserId"
			FROM "#message" m, "#keylist" k
			WHERE m."ReadAccessKeylistId" = k."KeylistId" AND m."MessageId" = ?',
			$this->messageId);
		$keys = array_map('intval', $keys);
		return $keys;
	}

	// Gets a list of all keys that allow reading this message, regarding access restrictions
	// imposed through assigned tags.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (array(int)) Array of keys that are allowed to read, or {{null}} if there is no access restriction
	//
	public function GetEffectiveReadAccessKeys()
	{
		$keylistIds = $this->db->QueryArray(
			'SELECT "ReadAccessKeylistId" AS "KeylistId"
			FROM "#message"
			WHERE "MessageId" = ? AND
				"ReadAccessKeylistId" IS NOT NULL
			UNION
			SELECT t."ReadAccessKeylistId" AS "KeylistId"
			FROM "#message" m
				JOIN "#message_revision_tag" mrt ON
					mrt."MessageId" = m."MessageId" AND
					mrt."RevisionNumber" = m."SearchRevision"
				JOIN "#tag" t USING ("TagId")
			WHERE m."MessageId" = ? AND
				t."ReadAccessKeylistId" IS NOT NULL',
			array($this->messageId, $this->messageId));
		$keylistIds = array_map('intval', $keylistIds);
		if (!count($keylistIds)) return null;   // No access restriction

#		$sql = '';
#		$parameters = array();
#		foreach ($keylistIds as $keylistId)
#		{
#			$sql .= ($sql ? ' INTERSECT ' : '') . 'SELECT "UserId" FROM "#keylist" WHERE "KeylistId" = ?';
#			$parameters[] = $keylistId;
#		}
#		$keys = $this->db->QueryArray($sql, $parameters);
#		$keys = array_map('intval', $keys);
#		return $keys;

		$keysArray = array();
		foreach ($keylistIds as $keylistId)
		{
			$keys = $this->db->QueryArray(
				'SELECT "UserId" FROM "#keylist" WHERE "KeylistId" = ?',
				$keylistId);
			$keys = array_map('intval', $keys);
			$keysArray[] = $keys;
		}

		// INTERSECT logic: (MySQL can't INTERSECT so we need to do that job)
		$intersectKeys = array();
		foreach ($keysArray[0] as $keyOfFirst)
		{
			$inAllOthers = true;
			for ($n = 1; $n < count($keysArray); $n++)
			{
				if (!in_array($keyOfFirst, $keysArray[$n]))
				{
					$inAllOthers = false;
					break;
				}
			}
			if ($inAllOthers)
				$intersectKeys[] = $keyOfFirst;
		}

		return $intersectKeys;
	}

	// Adds a key to this message that allows reading it.
	//
	// key = (int/UnbUser) Key to add
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::AddKeylistKey()
	//
	public function AddReadAccessKey($key)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to add read access keys.');

		self::ClearAccessCache();
		$this->db->AddKeylistKey('#message', 'ReadAccessKeylistId', 'MessageId', $this->messageId, $key);
	}

	// Removes a key from this message that allows reading it.
	//
	// key = (int/UnbUser) Key to remove
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::RemoveKeylistKey()
	//
	public function RemoveReadAccessKey($key)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to remove read access keys.');

		self::ClearAccessCache();
		$this->db->RemoveKeylistKey('#message', 'ReadAccessKeylistId', 'MessageId', $this->messageId, $key);
	}

	// Sets all keys for this message that allow reading it. This removes any keys set before.
	//
	// keys = (array(int/UnbUser)) Keys to set
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::ResetKeylistKeys()
	//
	public function SetReadAccessKeys($keys)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to set read access keys.');

		$this->db->ResetKeylistKeys('#message', 'ReadAccessKeylistId', 'MessageId', $this->messageId, $keys);
	}

	// Gets a list of all keys that allow altering this message.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (array(int)) Keys array
	//
	public function GetAlterAccessKeys()
	{
		$keys = $this->db->QueryArray(
			'SELECT k."UserId"
			FROM "#message" m, "#keylist" k
			WHERE m."AlterAccessKeylistId" = k."KeylistId" AND m."MessageId" = ?',
			$this->messageId);
		$keys = array_map('intval', $keys);
		return $keys;
	}

	// Gets a list of all keys that allow altering this message, regarding access restrictions
	// imposed through assigned tags.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (array(int)) Array of keys that are allowed to alter, or {{null}} if there is no access restriction
	//
	public function GetAllAlterAccessKeys()
	{
		$keylistIds = $this->db->QueryArray(
			'SELECT "AlterAccessKeylistId" AS "KeylistId"
			FROM "#message"
			WHERE "MessageId" = ? AND
				"AlterAccessKeylistId" IS NOT NULL
			UNION
			SELECT t."UseAccessKeylistId" AS "KeylistId"
			FROM "#message" m
				JOIN "#message_revision_tag" mrt ON
					mrt."MessageId" = m."MessageId" AND
					mrt."RevisionNumber" = m."SearchRevision"
				JOIN "#tag" t USING ("TagId")
			WHERE m."MessageId" = ? AND
				t."UseAccessKeylistId" IS NOT NULL',
			array($this->messageId, $this->messageId));
		$keylistIds = array_map('intval', $keylistIds);
		if (!count($keylistIds)) return null;   // No access restriction

		$keysArray = array();
		foreach ($keylistIds as $keylistId)
		{
			$keys = $this->db->QueryArray(
				'SELECT "UserId" FROM "#keylist" WHERE "KeylistId" = ?',
				$keylistId);
			$keys = array_map('intval', $keys);
			$keysArray[] = $keys;
		}

		// INTERSECT logic: (MySQL can't INTERSECT so we need to do that job)
		$intersectKeys = array();
		foreach ($keysArray[0] as $keyOfFirst)
		{
			$inAllOthers = true;
			for ($n = 1; $n < count($keysArray); $n++)
			{
				if (!in_array($keyOfFirst, $keysArray[$n]))
				{
					$inAllOthers = false;
					break;
				}
			}
			if ($inAllOthers)
				$intersectKeys[] = $keyOfFirst;
		}

		return $intersectKeys;
	}

	// Adds a key to this message that allows altering it.
	//
	// key = (int/UnbUser) Key to add
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::AddKeylistKey()
	//
	public function AddAlterAccessKey($key)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to add alter access keys.');

		self::ClearAccessCache();
		$this->db->AddKeylistKey('#message', 'AlterAccessKeylistId', 'MessageId', $this->messageId, $key);
	}

	// Removes a key from this message that allows altering it.
	//
	// key = (int/UnbUser) Key to remove
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::RemoveKeylistKey()
	//
	public function RemoveAlterAccessKey($key)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to remove alter access keys.');

		self::ClearAccessCache();
		$this->db->RemoveKeylistKey('#message', 'AlterAccessKeylistId', 'MessageId', $this->messageId, $key);
	}

	// Sets all keys for this message that allow altering it. This removes any keys set before.
	//
	// keys = (array(int/UnbUser)) Keys to set
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::ResetKeylistKeys()
	//
	public function SetAlterAccessKeys($keys)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to set alter access keys.');

		$this->db->ResetKeylistKeys('#message', 'AlterAccessKeylistId', 'MessageId', $this->messageId, $keys);
	}

	// Gets a list of all keys that allow replying to this message.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (array(int)) Keys array
	//
	public function GetReplyAccessKeys()
	{
		$keys = $this->db->QueryArray(
			'SELECT k."UserId"
			FROM "#message" m, "#keylist" k
			WHERE m."ReplyAccessKeylistId" = k."KeylistId" AND m."MessageId" = ?',
			$this->messageId);
		$keys = array_map('intval', $keys);
		return $keys;
	}

	// Adds a key to this message that allows replying to it.
	//
	// key = (int/UnbUser) Key to add
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::AddKeylistKey()
	//
	public function AddReplyAccessKey($key)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to add reply access keys.');
		// TODO: A moderator should be able to do this, too.

		self::ClearAccessCache();
		$this->db->AddKeylistKey('#message', 'ReplyAccessKeylistId', 'MessageId', $this->messageId, $key);
	}

	// Removes a key from this message that allows replying to it.
	//
	// key = (int/UnbUser) Key to remove
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::RemoveKeylistKey()
	//
	public function RemoveReplyAccessKey($key)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to remove reply access keys.');
		// TODO: A moderator should be able to do this, too.

		self::ClearAccessCache();
		$this->db->RemoveKeylistKey('#message', 'ReplyAccessKeylistId', 'MessageId', $this->messageId, $key);
	}

	// Sets all keys for this message that allow replying to it. This removes any keys set before.
	//
	// keys = (array(int/UnbUser)) Keys to set
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no administrator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with administrator privilege
	//
	// See UnbDatabase::ResetKeylistKeys()
	//
	public function SetReplyAccessKeys($keys)
	{
		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsAdministrator())
			throw new UnbSecurityException('You must be logged in as this user or need the administrator privilege to set reply access keys.');
		// TODO: A moderator should be able to do this, too.

		$this->db->ResetKeylistKeys('#message', 'ReplyAccessKeylistId', 'MessageId', $this->messageId, $keys);
	}

	// Gets the message's moderation flags.
	//
	// Returns: (int) Moderation flags
	//
	public function GetModerationFlags()
	{
		return $this->moderationFlags;
	}

	// Sets the message's moderation flags.
	//
	// moderationFlags = (int) New moderation flags
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetModerationFlags($moderationFlags)
	{
		if (!is_int($moderationFlags))
			throw new InvalidArgumentException('Invalid argument type: $moderationFlags');
		if ($moderationFlags < 0)
			throw new InvalidArgumentException('Argument out of range: $moderationFlags');

		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You need the moderator privilege to set moderation flags.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "ModerationFlags" = ? WHERE "MessageId" = ?',
				array($moderationFlags, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->moderationFlags = $moderationFlags;
	}

	// Gets the time when a user last began editing this message.
	//
	// Returns: (int) UNIX timestamp
	//
	public function GetLastBeginEdit()
	{
		return $this->lastBeginEdit;
	}

	// Sets the time when a user last began editing this message.
	//
	// You cannot set a specific time but rather indicate whether you have started editing now or
	// cancelled editing.
	//
	// editing = (bool) {{true}}: Indicate start of editing, sets time to now.
	//                  {{false}}: Indicate cancelled/ended editing, sets time to 0.
	//
	// Throws {{UnbDatabaseException}}
	//
	public function SetLastBeginEdit($editing)
	{
		$lastBeginEdit = $editing ? time() : 0;

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "LastBeginEdit" = ? WHERE "MessageId" = ?',
				array(UnbDatabase::ConvertDateTime($lastBeginEdit), $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->lastBeginEdit = $lastBeginEdit;
	}

	// Gets the content of this message that shall be regarded for searching.
	//
	// Returns: (string)
	//
	public function GetSearchOptimisedContent()
	{
		return $this->searchOptimisedContent;
	}

	// Sets the content of this message that shall be regarded for searching.
	//
	// content = (string) New content or {{null}} (max. 16,777,215 characters)
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{UnbDatabaseException}}
	//
	public function SetSearchOptimisedContent($content)
	{
		if (isset($content))
		{
			if (!is_string($content))
				throw new InvalidArgumentException('Invalid argument type: $content');
			$content = trim($content);
			if (mb_strlen($content) > 16777215)
				throw new LengthException('[#3] Content is too long');
			if (!strlen($content))
				$content = null;
		}

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "SearchOptimisedContent" = ? WHERE "MessageId" = ?',
				array($content, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->searchOptimisedContent = $content;
	}

	// Gets the subject of this message that shall be regarded for searching.
	//
	// Returns: (string)
	//
	public function GetSearchOptimisedSubject()
	{
		return $this->searchOptimisedSubject;
	}

	// Sets the subject of this message that shall be regarded for searching.
	//
	// subject = (string) New subject or {{null}} (max. 255 characters)
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{UnbDatabaseException}}
	//
	public function SetSearchOptimisedSubject($subject)
	{
		if (isset($subject))
		{
			if (!is_string($subject))
				throw new InvalidArgumentException('Invalid argument type: $subject');
			$subject = trim($subject);
			if (mb_strlen($subject) > 255)
				throw new LengthException('[#4] Subject is too long');
			if (!strlen($subject))
				$subject = null;
		}

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message" SET "SearchOptimisedSubject" = ? WHERE "MessageId" = ?',
				array($subject, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->searchOptimisedSubject = $subject;
	}

	// Gets the revision number that shall be regarded for searching.
	//
	// Returns: (int)
	//
	public function GetSearchRevisionNumber()
	{
		return $this->searchRevision;
	}

	// Gets the revision object that shall be regarded for searching.
	//
	// contentLevel = (int) See UnbMessageRevision::FindByMessageId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessageRevision object if successful, {{null}} otherwise
	//
	// See UnbMessage::GetRevision()
	//
	public function GetSearchRevision($contentLevel = null, $allowFallback = false)
	{
		return $this->GetRevision($this->searchRevision, $contentLevel, $allowFallback);
	}

	// Sets the revision number that shall be regarded for searching.
	//
	// revisionNumber = (int) New revision number
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{UnbDatabaseException}}
	//
	public function SetSearchRevisionNumber($revisionNumber)
	{
		if (isset($revisionNumber))
		{
			if (!is_int($revisionNumber))
				throw new InvalidArgumentException('Invalid argument type: $revisionNumber');
			if ($revisionNumber <= 0)
				throw new InvalidArgumentException('Argument out of range: $revisionNumber');
		}

		$this->db->StartTransaction();
		try
		{
			// TODO: Check if the revision number is valid - can this be done at the database level? (Foreign key constraint, see setup.sql)

			$this->db->Execute(
				'UPDATE "#message" SET "SearchRevision" = ? WHERE "MessageId" = ?',
				array($revisionNumber, $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
		$this->searchRevision = intval($revisionNumber);
	}

	// Deletes this instance of the message with all its references.
	//
	// Throws {{UnbSecurityException}} If not logged in or no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function Delete()
	{
		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to delete a message.');

		$this->db->StartTransaction();
		try
		{
			// Deleting messages may leave keylists unreferenced. Delete them now.
			$this->db->ResetKeylistKeys('#message', 'ReadAccessKeylistId', 'MessageId', $this->messageId, null);
			$this->db->ResetKeylistKeys('#message', 'AlterAccessKeylistId', 'MessageId', $this->messageId, null);
			$this->db->ResetKeylistKeys('#message', 'ReplyAccessKeylistId', 'MessageId', $this->messageId, null);

			// All other references are automatically resolved by the database foreign key contraints. :)

			$this->db->Execute('DELETE FROM "#message" WHERE "MessageId" = ?', $this->messageId);

			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Sets the search-optimised fields for content, subject and revision number to the values
	// from the latest approved message revision.
	//
	// messageId = (int) Message ID
	//
	// If there is no approved message revision, the latest of all revisions is used instead.
	//
	// Throws {{UnbDatabaseException}}
	// Throws {{InvalidArgumentException}}
	//
	public static function UpdateSearchFields($messageId)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction(true);
		try
		{
			$revisionNumber = $db->QueryValue(
				'SELECT MAX("RevisionNumber")
				FROM "#message_revision"
				WHERE "MessageId" = ? AND "ModerationState" = ' . UnbMessageRevisionModerationState::Approved . '
				FOR UPDATE',
				$messageId);
			if (!isset($revisionNumber))
				// No approved revision found, now accept any moderation state
				$revisionNumber = $db->QueryValue(
					'SELECT MAX("RevisionNumber")
					FROM "#message_revision"
					WHERE "MessageId" = ?
					FOR UPDATE',
					$messageId);

			$rec = $db->QueryRecord(
				'SELECT "Subject", "Content", "ModerationState"
				FROM "#message_revision"
				WHERE "MessageId" = ? AND "RevisionNumber" = ?',
				array($messageId, $revisionNumber));

			$content = null;
			$subject = null;
			if (isset($rec) && $rec['ModerationState'] == UnbMessageRevisionModerationState::Approved)
			{
				// TODO: Convert original values for search-optimised use
				#$content = $rec['Content'];
				#$subject = $rec['Subject'];
			}

			$db->Execute(
				'UPDATE "#message"
				SET "SearchOptimisedContent" = ?, "SearchOptimisedSubject" = ?, "SearchRevision" = ?
				WHERE "MessageId" = ?',
				array($content, $subject, $revisionNumber, $messageId));
			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		#$this->searchRevision = $revisionNumber;
	}

	// Gets all additional message IDs that this message is referencing.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (array(int)) Reference message IDs
	//
	public function GetMoreReferences()
	{
		// Sort referenced messages by the time of their first revision.
		// To sort by the last revision (no access check!) replace MIN(CreatedTime) function by MAX(CreatedTime).
		$sql = 'SELECT mr."Reference", MIN(mrev."CreatedTime") AS ct
			FROM "#message_reference" mr,
				"#message" m
				JOIN "#message_revision" mrev USING ("MessageId")
			WHERE mr."Reference" = m."MessageId" AND
				mr."MessageId" = ?
			GROUP BY mr."Reference"
			ORDER BY ct';
		$parameters = $this->messageId;

		$refs = $this->db->QueryArray($sql, $parameters);
		$refs = array_map('intval', $refs);
		return $refs;
	}

	// Adds a message reference to this message.
	//
	// reference = (int/UnbMessage) Reference to add
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with moderator privilege
	//
	public function AddReference($reference)
	{
		if ($reference instanceof UnbMessage)
			$reference = $reference->GetMessageId();
		if (!is_int($reference))
			throw new InvalidArgumentException('Invalid argument type: $reference');
		if ($reference <= 0)
			throw new InvalidArgumentException('Argument out of range: $reference');

		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsModerator())
			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to add a reference.');

		$this->db->StartTransaction();
		try
		{
			try
			{
				$this->db->Execute(
					'INSERT INTO "#message_reference" ("MessageId", "Reference") VALUES (?, ?)',
					array($this->messageId, $reference));
			}
			catch (UnbDatabaseException $ex)
			{
				// If the entry already exists, it is unchanged.
				if (!$this->db->IsConstraintError()) throw $ex;
			}
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Removes a message reference from this message.
	//
	// reference = (int/UnbMessage) Reference to remove
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the message owner/no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the message owner or with moderator privilege
	//
	public function RemoveReference($reference)
	{
		if ($reference instanceof UnbMessage)
			$reference = $reference->GetMessageId();
		if (!is_int($reference))
			throw new InvalidArgumentException('Invalid argument type: $reference');
		if ($reference <= 0)
			throw new InvalidArgumentException('Argument out of range: $reference');

		if (UnbSession::GetSessionUserId() != $this->ownerUserId && !UnbSession::IsModerator())
			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to remove a reference.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'DELETE FROM "#message_reference" WHERE "MessageId" = ? AND "Reference" = ?',
				array($this->messageId, $reference));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets the last-referenced discussion entry point's message ID.
	//
	// If this message is already a discussion entry point or has no primary reference, its message
	// ID is returned. Otherwise, each primary reference will be followed until a message either
	// marked as entry point or with no primary reference is found.
	//
	// Returns (int) Message ID of the last-referenced entry point
	//
	// Throws {{UnbDatabaseException}} If a database error occured or if a primary reference loop was detected.
	//
	public function GetRefDEP()
	{
		if ($this->entryPoint) return $this->messageId;
		if (!$this->primaryReference) return $this->messageId;

		$messageId = $this->primaryReference;
		$seenMessageIds = array();   // Infinite loop protection
		while (true)
		{
			$seenMessageIds[] = $messageId;
			$rec = $this->db->QueryRecord(
				'SELECT "EntryPoint", "PrimaryReference" FROM "#message" WHERE "MessageId" = ?',
				$messageId);
			if ($rec['EntryPoint']) return $messageId;
			$primaryReference = intval($rec['PrimaryReference']);
			if (!$primaryReference) return $messageId;
			$messageId = $primaryReference;

			if (in_array($messageId, $seenMessageIds))
				throw new UnbDatabaseException('Recursive primary message references detected, starting at ' . $this->messageId);
		}
	}

	// Gets the latest message from all messages that are directly or indirectly referencing to
	// this message through their primary reference.
	//
	// Returns (UnbMessage) The message instance of the latest reply according to its CreatedTime
	//
	// TODO: Primary reference loop detection, throw UnbDatabaseException like in GetRefDEP()
	//
	public function GetLatestReply()
	{
		$latestRecord = array();
		$latestTime = '';
		$refIds = array($this->messageId);
		while (true)
		{
			$recs = $this->db->Query('SELECT * FROM "#message" WHERE "PrimaryReference" IN (' . join(', ', $refIds) . ')');
			if (!$recs) break;
			$refIds = array();
			foreach ($recs as $rec)
			{
				if ($rec['CreatedTime'] > $latestTime)
				{
					$latestTime = $rec['CreatedTime'];
					$latestRecord = $rec;
				}
				$refIds[] = intval($rec['MessageId']);
			}
		}
		if ($latestRecord)
			return self::FromRecord($latestRecord);
		// No reply found
		return null;
	}

	// Gets the number of messages that are directly or indirectly referencing to this message
	// through their primary reference.
	//
	// Returns (int) The number of all replies
	//
	// TODO: Primary reference loop detection, throw UnbDatabaseException like in GetRefDEP()
	//
	public function CountReplies()
	{
		$count = 0;
		$refIds = array($this->messageId);
		while (true)
		{
			$recs = $this->db->Query('SELECT "MessageId" FROM "#message" WHERE "PrimaryReference" IN (' . join(', ', $refIds) . ')');
			if (!$recs) break;
			$refIds = array();
			foreach ($recs as $rec)
			{
				$count++;
				$refIds[] = intval($rec['MessageId']);
			}
		}
		return $count;
	}

#	// Gets the time when the message was first created.
#	//
#	// This time is read from the first revision's creation timestamp.
#	//
#	// Throws {{UnbDatabaseException}}
#	//
#	// Returns: (int) UNIX timestamp
#	//
#	public function GetCreatedTime()
#	{
#		return UnbMessageRevision::FirstCreatedTime($this->messageId);
#	}
# TODO: See new GetCreatedTime() method in this class

	// Adds a new revision to this message.
	//
	// subject = (string) Subject of the new revision (max. 255 characters)
	// content = (string) Content (max. 16,777,215 characters)
	// summary = (string) Summary of the changes in the new revivion to the previous one (max. 255 characters)
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessageRevision object
	//
	// Privileges: See UnbMessageRevision::Create()
	//
	// See UnbMessageRevision::Create()
	//
	public function AddRevision($subject, $content, $summary = '')
	{
		return UnbMessageRevision::Create($this->messageId, $subject, $content, $summary);
	}

	// Gets a specific revision of this message by its revision number.
	//
	// revisionNumber = (int) Revision number to get
	// contentLevel = (int) See UnbMessageRevision::FindByMessageId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessageRevision object if successful, {{null}} otherwise
	//
	// See UnbMessageRevision::FindByMessageId()
	//
	public function GetRevision($revisionNumber, $contentLevel = null, $allowFallback = false)
	{
		return UnbMessageRevision::FindByMessageId($this->messageId, $revisionNumber, $contentLevel, $allowFallback);
	}

	// Gets the latest accessible revision of this message by its moderation state.
	//
	// contentLevel = (int) See UnbMessageRevision::FindByMessageId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessageRevision object if successful, {{null}} otherwise
	//
	// See UnbMessageRevision::FindLatestByMessageId()
	//
	public function GetLatestRevision($contentLevel = null, $allowFallback = false)
	{
		return UnbMessageRevision::FindLatestByMessageId($this->messageId, $contentLevel, $allowFallback);
	}

	// Gets all revisions of this message, with no content.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: Array of UnbMessageRevision objects, if successful, {{null}} otherwise.
	//
	// See UnbMessageRevision::FindAllByMessageId()
	//
	public function GetAllRevisions()
	{
		return UnbMessageRevision::FindAllByMessageId($this->messageId);
	}

	// Gets this message's overall rating.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: An array(Number of ratings, Rating weight, Weighted rating average), or {{null}} if failed.
	//
	public function GetRating()
	{
		$rec = $this->db->QueryRecord(
			'SELECT COUNT(*) AS "Count", SUM("Weight") AS "TotalWeight", SUM("Rating" * "Weight") AS "WeightedSum"
			FROM "#message_rating"
			WHERE "MessageId" = ?',
			$this->messageId);
		if (!$rec) return null;
		// Keep the total weight to 100 at least to weaken few ratings
		$totalWeight = max(100.0, floatval($rec['TotalWeight']));
		return array(
			intval($rec['Count']),
			intval($rec['TotalWeight']),
			floatval($rec['WeightedSum']) / $totalWeight);
	}

	// Gets the session user's rating for this message.
	//
	// Throws {{UnbSecurityException}} If not logged in.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: The rating value, or {{null}} if there is no rating from the user.
	//
	// Privileges: Must be logged in (obviously)
	//
	public function GetMyRating()
	{
		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You must be logged in to get your rating.');

		$r = $this->db->QueryValue(
			'SELECT "Rating" FROM "#message_rating" WHERE "UserId" = ? AND "MessageId" = ?',
			array(UnbSession::GetSessionUserId(), $this->messageId));
		if (!isset($r)) return null;
		return intval($r);
	}

	// Sets the user's rating for this message.
	//
	// rating = (int) Rating value, where 0 is a neutral rating, positive values are good, negative are bad.
	//                Valid ratings are in the range from -10 to 10 (inclusive).
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or rating own messages.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: {{true}}, if successful, {{false}} otherwise.
	//
	// Privileges: Must be logged in
	//
	public function SetRating($rating)
	{
		if (!is_int($rating))
			throw new InvalidArgumentException('Invalid argument type: $rating');
		if (abs($rating) > 10)
			throw new InvalidArgumentException('Argument out of range: $rating');

		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You must be logged in to rate a message.');

		// Disallow users rating their own messages
		if (UnbSession::GetSessionUserId() == $this->ownerUserId)
			throw new UnbSecurityException('You cannot rate your own messages.');

		// Get the user's own rating to calculate his rating weight for this message
		$userRating = UnbSession::GetSessionUser()->GetMessagesRating();
		if (!isset($userRating)) return false;
		$myWeight = $userRating[2] + 10;   // Normalise value from [-10,10] to [0,20]

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message_rating" SET "Rating" = ?, "Weight" = ? WHERE "UserId" = ? AND "MessageId" = ?',
				array($rating, $myWeight, UnbSession::GetSessionUserId(), $this->messageId));
			if ($this->db->GetRowsAffected() == 0)
			{
				// The entry doesn't exist yet, so insert it.
				try
				{
					$this->db->Execute(
						'INSERT INTO "#message_rating" ("UserId", "MessageId", "Rating", "Weight") VALUES (?, ?, ?, ?)',
						array(UnbSession::GetSessionUserId(), $this->messageId, $rating, $myWeight));
				}
				catch (UnbDatabaseException $ex)
				{
					// If the entry already exists, it is unchanged.
					// GetRowsAffected() has returned 0 in this case although the record has been found.
					// But it has not been altered because it is unchanged.
					if (!$this->db->IsConstraintError()) throw $ex;
				}
			}
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Clears the user's rating for this message.
	//
	// Throws {{UnbSecurityException}} If not logged in or rating own messages.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: {{true}}, if successful, {{false}} otherwise.
	//
	// Privileges: Must be logged in
	//
	public function ClearRating()
	{
		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You must be logged in to rate a message.');

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'DELETE FROM "#message_rating" WHERE "UserId" = ? AND "MessageId" = ?',
				array(UnbSession::GetSessionUserId(), $this->messageId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets a value indicating whether reading this message is allowed to the current session user.
	//
	// This does not regard moderation flags.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (bool)
	//
	public function IsReadAllowed()
	{
		return self::IsReadAllowedById($this->messageId);
	}

	// Gets a value indicating whether reading a message specified by its message ID is allowed to
	// the current session user.
	//
	// This does not regard moderation flags. It also won't give any exceptional access, like for
	// administrators or the message owner. These exceptions need to be handled by the code that
	// calls this method, according to the UNB access specification.
	//
	// messageId = (int/UnbMessage) Message or its ID to check.
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (bool)
	//
	public static function IsReadAllowedById($messageId)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');

		if (array_key_exists($messageId, self::$readAccessCache))
			return self::$readAccessCache[$messageId];

		$sessionUserId = UnbSession::GetSessionUserId();

		$sql = 'SELECT
			(EXISTS
				-- Message has no read access keylist
				(SELECT 1
				FROM "#message" m
				WHERE m."MessageId" = ? AND
					m."ReadAccessKeylistId" IS NULL)
			OR EXISTS
				-- Message\'s read access keylist contains personal or additional user key
				(SELECT 1
				FROM "#message" m, "#keylist" mk
				WHERE mk."KeylistId" = m."ReadAccessKeylistId" AND
					m."MessageId" = ? AND
					(mk."UserId" = ? OR
						mk."UserId" IN
							(SELECT uk."UserId"
							FROM "#user" u, "#keylist" uk
							WHERE uk."KeylistId" = u."AdditionalKeylistId" AND
								u."UserId" = ?))))
			AND NOT EXISTS
				-- There must not be any tag that denies access
				(SELECT "Allowed"
				FROM
					(SELECT
						EXISTS
							-- Tag\'s read access keylist contains personal or additional user key
							(SELECT 1
							FROM "#keylist" tk
							WHERE tk."KeylistId" = t."ReadAccessKeylistId" AND
								(tk."UserId" = ? OR
									tk."UserId" IN
										(SELECT uk."UserId"
										FROM "#user" u, "#keylist" uk
										WHERE uk."KeylistId" = u."AdditionalKeylistId" AND
											u."UserId" = ?)))
						AS "Allowed"
					FROM "#message" m
						JOIN "#message_revision_tag" mrt ON
							mrt."MessageId" = m."MessageId" AND
							mrt."RevisionNumber" = m."SearchRevision"
						JOIN "#tag" t USING ("TagId")
					WHERE m."MessageId" = ? AND
						-- Only regard tags with read access keylist (only they can fail the access test)
						t."ReadAccessKeylistId" IS NOT NULL) AS foo
				WHERE NOT "Allowed")';
		// Notes on this query:
		// 1. PostgreSQL requires the "Allowed" thing to be moved to its own subquery
		//    because it can't select on computed column in a HAVING clause.
		// 2. The HAVING clause would be required because you cannot use computed columns
		//    in a WHERE clause since they might not be computed yet when selecting rows.
		$parameters = array($messageId, $messageId, $sessionUserId, $sessionUserId, $sessionUserId, $sessionUserId, $messageId);

		$db = UnbDatabase::GetInstance();
		$val = $db->QueryValue($sql, $parameters);
		$val = $val ? true : false;
		self::$readAccessCache[$messageId] = $val;
		return $val;
	}

	// Remember that the specified message IDs are readable by the current session user.
	//
	// This is only an optimisation in cooperation with the message search that has already
	// determined that all search results are readable. Using this class' private cache, this
	// doesn't need to be determined again. So actually, this method is for internal use only!
	//
	// messageIds = (array) Message IDs to remember as readable
	//
	public static function SetReadAllowedIds($messageIds)
	{
		foreach ($messageIds as $messageId)
		{
			if (!array_key_exists($messageId, self::$readAccessCache))
				self::$readAccessCache[$messageId] = true;
		}
	}

	// Gets a value indicating whether altering this message is allowed to the current session user.
	//
	// This does not regard moderation flags.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (bool)
	//
	public function IsAlterAllowed()
	{
		return self::IsAlterAllowedById($this->messageId);
	}

	// Gets a value indicating whether altering a message specified by its message ID is allowed
	// to the current session user.
	//
	// messageId = (int/UnbMessage) Message or its ID to check.
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (bool)
	//
	public static function IsAlterAllowedById($messageId)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');

		if (array_key_exists($messageId, self::$alterAccessCache))
			return self::$alterAccessCache[$messageId];

		$sessionUserId = UnbSession::GetSessionUserId();
		if (!$sessionUserId) return false;   // Must be logged in to alter a message

		$sql = 'SELECT
			(EXISTS
				-- Message has no alter access keylist
				(SELECT 1
				FROM "#message" m
				WHERE m."MessageId" = ? AND
					m."AlterAccessKeylistId" IS NULL)
			OR EXISTS
				-- Message\'s alter access keylist contains personal or additional user key
				(SELECT 1
				FROM "#message" m, "#keylist" mk
				WHERE mk."KeylistId" = m."AlterAccessKeylistId" AND
					m."MessageId" = ? AND
					(mk."UserId" = ? OR
						mk."UserId" IN
							(SELECT uk."UserId"
							FROM "#user" u, "#keylist" uk
							WHERE uk."KeylistId" = u."AdditionalKeylistId" AND
								u."UserId" = ?))))
			AND NOT EXISTS
				-- There must not be any tag that denies access
				(SELECT "Allowed"
				FROM
					(SELECT
						EXISTS
							-- Tag\'s use access keylist contains personal or additional user key
							(SELECT 1
							FROM "#keylist" tk
							WHERE tk."KeylistId" = t."UseAccessKeylistId" AND
								(tk."UserId" = ? OR
									tk."UserId" IN
										(SELECT uk."UserId"
										FROM "#user" u, "#keylist" uk
										WHERE uk."KeylistId" = u."AdditionalKeylistId" AND
											u."UserId" = ?)))
						AS "Allowed"
					FROM "#message" m
						JOIN "#message_revision_tag" mrt ON
							mrt."MessageId" = m."MessageId" AND
							mrt."RevisionNumber" = m."SearchRevision"
						JOIN "#tag" t USING ("TagId")
					WHERE m."MessageId" = ? AND
						-- Only regard tags with use access keylist (only they can fail the access test)
						t."UseAccessKeylistId" IS NOT NULL) AS foo
				WHERE NOT "Allowed")';
		// Notes on this query: see IsReadAllowedById() above
		$parameters = array($messageId, $messageId, $sessionUserId, $sessionUserId, $sessionUserId, $sessionUserId, $messageId);

		$db = UnbDatabase::GetInstance();
		$val = $db->QueryValue($sql, $parameters);
		$val = $val ? true : false;
		self::$alterAccessCache[$messageId] = $val;
		return $val;
	}

	// Gets a value indicating whether replying to this message is allowed to the current session user.
	//
	// This does not regard moderation flags.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (bool)
	//
	public function IsReplyAllowed()
	{
		return self::IsReplyAllowedById($this->messageId);
	}

	// Gets a value indicating whether replying to a message specified by its message ID is
	// allowed to the current session user.
	//
	// messageId = (int/UnbMessage) Message or its ID to check.
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{UnbDatabaseException}}
	//
	// Returns: (bool)
	//
	public static function IsReplyAllowedById($messageId)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');

		if (array_key_exists($messageId, self::$replyAccessCache))
			return self::$replyAccessCache[$messageId];

		$sessionUserId = UnbSession::GetSessionUserId();

		$sql = 'SELECT
			(EXISTS
				-- Message has no reply access keylist
				(SELECT 1
				FROM "#message" m
				WHERE m."MessageId" = ? AND
					m."ReplyAccessKeylistId" IS NULL)
			OR EXISTS
				-- Message\'s reply access keylist contains personal or additional user key
				(SELECT 1
				FROM "#message" m, "#keylist" mk
				WHERE mk."KeylistId" = m."ReplyAccessKeylistId" AND
					m."MessageId" = ? AND
					(mk."UserId" = ? OR
						mk."UserId" IN
							(SELECT uk."UserId"
							FROM "#user" u, "#keylist" uk
							WHERE uk."KeylistId" = u."AdditionalKeylistId" AND
								u."UserId" = ?))))';
		$parameters = array($messageId, $messageId, $sessionUserId, $sessionUserId);

		$db = UnbDatabase::GetInstance();
		$val = $db->QueryValue($sql, $parameters);
		$val = $val ? true : false;
		self::$replyAccessCache[$messageId] = $val;
		return $val;
	}

	// Clears this message's internal cache of access key lists.
	//
	public static function ClearAccessCache()
	{
		self::$readAccessCache = array();
		self::$alterAccessCache = array();
		self::$replyAccessCache = array();
	}

} // class UnbMessage

?>