<?php
// UNB Components
// Copyright 2007-2010 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UnbMessageRevision class.

if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';
if (!class_exists('UnbMessage', false)) require_once dirname(__FILE__) . '/message.class.php';
if (!class_exists('UnbTag', false)) require_once dirname(__FILE__) . '/tag.class.php';
if (!class_exists('UnbSession', false)) require_once dirname(__FILE__) . '/session.class.php';
if (!class_exists('UnbConfiguration', false)) require_once dirname(__FILE__) . '/configuration.class.php';
if (!class_exists('UnbSecurityException', false)) require_once dirname(__FILE__) . '/unb.class.php';

// Enumeration of moderation states for message revisions.
//
class UnbMessageRevisionModerationState
{
	// This revision is waiting to be approved
	const WaitingForApproval = 0;
	// This message revision has been approved
	const Approved = 1;
	// This message revision has been approved but is now locked
	const Locked = 2;
}

// Enumeration of message revision retrieval levels regarding content fields (content level).
//
class UnbMessageRevisionContentLevel
{
	// Don't retrieve any content
	const NoContent = 0;
	// Only retrieve the subject and summary fields
	const Subject = 1;
	// Also retrieve the source content field
	const Content = 2;
	// Retrieve both the source content and the cached HTML content
	const HtmlContent = 3;
}

// The message revision class represents a single message revision in the UNB system and provides
// methods to find and manipulate message revision instances.
//
// It depends on the UnbDatabase class for a database connection. It also uses the class UnbTag to
// manage message revision tags.
//
class UnbMessageRevision extends UnbEntity
{
	// ID of the message that this revision belongs to
	private $messageId = 0;
	// Number of the revision, started at 1, increaded by 1 with every new revision
	private $revisionNumber = 0;
	// Time when this revision was created
	private $createdTime = 0;
	// User who created this revision
	private $author = 0;
	// Subject of the message (revision)
	private $subject = '';
	// Content
	private $content = '';
	// Cached HTML output of the parsed content, used to accelerate displaying of the message (revision)
	private $htmlContent = '';
	// Summary of the changes in this revivion to the previous one
	private $summary = '';
	// Moderation state of this single revision (see UnbMessageRevision::MS_* constants)
	private $moderationState = 0;
	// Does this message revision have attached data?
	private $hasData = false;

	// Cached author UnbUser object
	private $authorUser = null;
	// Remembers which level of content was requested for this revision
	private $contentLevel = 0;

	// Cached UnbDatabase instance
	private $db;

	private function __construct()
	{
		$this->db = UnbDatabase::GetInstance();
	}

	// Creates a new message revision instance for a message.
	//
	// The newly created message revision instance is also stored in the database. If the session
	// user is a moderator, the new message revision is automatically approved.
	//
	// messageId = (int/UnbMessage) Message to add the revision to
	// subject = (string) Subject of the message's first revision or {{null}} (max. 255 characters)
	// content = (string) Content of the message's first revision or {{null}} (max. 16,777,215 characters)
	// summary = (string) Summary of the changes in the new revivion to the previous one or {{null}} (max. 255 characters)
	// data = (string) Attached file data contents or {{null}}
	// dataFileName = (string) Local file name for attached file data or {{null}} (do not use together with {{date}})
	// writeToFile = (bool) Write the attached file data into a file instead of the database or {{null}} for automatic decision
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{Exception}} If the input file could not be read or copied.
	// Throws {{UnbSecurityException}} If not logged in or no read/alter access on the message.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: The newly created UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Must be logged in and need read and alter access on the message
	//
	public static function Create($messageId, $subject, $content, $summary = null, $data = null, $dataFileName = null, $writeToFile = null)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		if (isset($subject))
		{
			if (!is_string($subject))
				throw new InvalidArgumentException('Invalid argument type: $subject');
			$subject = trim($subject);
			if (mb_strlen($subject) > 255)
				throw new LengthException('[#4] Subject is too long');
			if (!strlen($subject)) $subject = null;
		}
		if (isset($content))
		{
			if (!is_string($content))
				throw new InvalidArgumentException('Invalid argument type: $content');
			// UNIX line endings
			$content = str_replace("\r", '', $content);
			// Trim all lines at their end
			$content = preg_replace('_[ \t]+(?=\n)_', '', $content);
			// Trim content at the end
			$content = rtrim($content);
			if (mb_strlen($content) > 16777215)
				throw new LengthException('[#3] Content is too long');
			if (!strlen(trim($content))) $content = null;
		}
		if (isset($summary))
		{
			if (!is_string($summary))
				throw new InvalidArgumentException('Invalid argument type: $summary');
			$summary = trim($summary);
			if (mb_strlen($summary) > 255)
				throw new LengthException('[#5] Summary is too long');
			if (!strlen($summary)) $summary = null;
		}
		if (isset($data))
		{
			if (!is_string($data))
				throw new InvalidArgumentException('Invalid argument type: $data');
		}
		if (isset($dataFileName))
		{
			if (!is_string($dataFileName))
				throw new InvalidArgumentException('Invalid argument type: $dataFileName');
			if (isset($data))
				throw new InvalidArgumentException('You cannot set $data and $dataFileName at the same time.');
			if (!strlen($dataFileName)) $dataFileName = null;
			if (isset($dataFileName) && !file_exists($dataFileName))
				throw new InvalidArgumentException('The specified file does not exist.');
		}

		if (!UnbSession::IsLoggedIn())
			throw new UnbSecurityException('You must be logged in to create a message revision.');

		$now = time();

		// Check for read and alter access on the message
		if (!UnbMessage::IsReadAllowedById($messageId))
			throw new UnbSecurityException('You are not allowed to read this message.');
		if (!UnbMessage::IsAlterAllowedById($messageId))
			throw new UnbSecurityException('You are not allowed to alter this message.');

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction(true);
		try
		{
			$revisionNumber = $db->QueryValue(
				'SELECT MAX("RevisionNumber") FROM "#message_revision" WHERE "MessageId" = ? FOR UPDATE',
				$messageId);
			if (!isset($revisionNumber))
				$revisionNumber = 0;
			$revisionNumber++;

			$fields = '';
			$values = '';
			$parameters = array(
				$messageId,
				$revisionNumber,
				UnbDatabase::ConvertDateTime($now),
				UnbSession::GetSessionUserId(),
				$subject,
				$content,
				$summary);

			$autoApprove = false;
			if (UnbSession::IsModerator())
				$autoApprove = true;
			if (UnbSession::IsTrusted() && UnbConfiguration::Get('moderation.approve-from-trusted', false))
				$autoApprove = true;
			if ($autoApprove)
			{
				// Moderators' new message (revisions) are always approved by default (self-approval)
				// TODO: Self-approval should be disabled for a specified list of "banned" users by option
				//       Use another system key for that option? Like UnbUserKey::Banned
				//       (Moderators cannot be banned - they must be excluded from this option)
				// TODO: This is the place to regard the EnforceApproval flag from referenced messages
				$fields .= ', "ModerationState"';
				$values .= ', ?';
				$parameters[] = UnbMessageRevisionModerationState::Approved;
			}

			// Handle attached file data
			if (isset($data) || isset($dataFileName))
			{
				$fields .= ', "HasData"';
				$values .= ', ?';
				$parameters[] = 1;

				if (!isset($writeToFile))
				{
					// No indication whether to save data in a file or in the database, ask our configuration
					$dbMaxSize = UnbConfiguration::Get('attachment.storage.database-max-size', -1);
					if ($dbMaxSize == -1)
					{
						// Always write to the database (default if unconfigured)
						$writeToFile = false;
					}
					else
					{
						// Write to the database up to a limit, larger data goes into a file
						if (is_string($data))
							$dataLength = strlen($data);
						else
							$dataLength = filesize($dataFileName);
						$writeToFile = $dataLength > $dbMaxSize;
					}
				}

				if ($writeToFile)
				{
					// Output is a file

					// Now write the new file
					$newFilename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $messageId . '_' . $revisionNumber;
					if (isset($dataFileName))
					{
						// Input is a file
						if (is_uploaded_file($dataFileName))
						{
							if (!move_uploaded_file($dataFileName, $newFilename))
								throw new Exception('[#6] The uploaded file could not be moved to its new location.');
						}
						else
						{
							if (!copy($dataFileName, $newFilename))
								throw new Exception('[#6] The input file could not be copied to its new location.');
						}
					}
					else
					{
						// Input is a string
						if (file_put_contents($newFilename, $data) === false)
							throw new Exception('[#7] The data could not be written to the local file.');
					}
				}
				else
				{
					// Output is database
					if (is_string($dataFileName))
					{
						// Input is a file
						$data = file_get_contents($dataFileName);
						if ($data === false)
							throw new Exception('[#8] The data could not be read from the input file.');
					}

					$fields .= ', "Data"';
					$values .= ', ?';
					$parameters[] = $data;
				}
			}

			$sql = 'INSERT INTO "#message_revision" ("MessageId", "RevisionNumber", "CreatedTime", "Author", "Subject", "Content", "Summary"' . $fields . ') VALUES (?, ?, ?, ?, ?, ?, ?' . $values . ')';
			$db->Execute($sql, $parameters);

			UnbMessage::UpdateSearchFields($messageId);

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		$messageRevision = new self();
		$messageRevision->messageId = $messageId;
		$messageRevision->revisionNumber = $revisionNumber;
		$messageRevision->createdTime = $now;
		$messageRevision->author = UnbSession::GetSessionUserId();
		$messageRevision->subject = $subject;
		$messageRevision->content = $content;
		$messageRevision->summary = $summary;
		$messageRevision->hasData = isset($data) || isset($dataFileName);
		return $messageRevision;
	}

	// Creates a new message revision instance for a message, based on this revision.
	//
	// The newly created message revision instance is also stored in the database. The same
	// conditions as for the static {{Create}} method apply. The data for one or multiple fields
	// of the new message revision can be copied from this revision by setting the respective
	// argument to {{true}}.
	//
	// This message revision must have been retrieved with content level {{UnbMessageRevisionContentLevel::Content}}.
	//
	// subject = (string) Subject of the message's first revision, {{true}} to copy or {{null}} (max. 255 characters)
	// content = (string) Content of the message's first revision, {{true}} to copy or {{null}} (max. 16,777,215 characters)
	// summary = (string) Summary of the changes in the new revivion to the previous one or {{null}} (max. 255 characters)
	// data = (string) Attached file data contents, {{true}} to copy or {{null}}
	// dataFileName = (string) Local file name for attached file data or {{null}} (do not use together with {{data}})
	// writeToFile = (bool) Write the attached file data into a file instead of the database or {{null}} for automatic decision
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{Exception}} If the input file could not be read or copied.
	// Throws {{UnbSecurityException}} If not logged in or no read/alter access on the message.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: The newly created UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Must be logged in and need read and alter access on the message
	//
	public function CreateCopy($subject, $content, $summary, $data, $dataFileName, $writeToFile)
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the content.');

		if ($data === true)
		{
			$data = $this->GetFileData(2);   // Only copy in-database data to this variable
			$dataFileName = $this->GetFileDataPath();   // Ignore possible local argument

			// NOTE:
			// MySQL doesn't currently support SELECTing from a table being UPDATEd. Sad but true.
			// So even if the data is copied from one revision to the new all inside the database,
			// it needs to be read into PHP and written again the regular way.
			// See http://dev.mysql.com/doc/refman/5.1/en/update.html (at the very bottom)
		}

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction(true);
		try
		{
			$newRevision = self::Create(
				$this->messageId,
				($subject === true ? $this->subject : $subject),
				($content === true ? $this->content : $content),
				$summary,   // Do not copy the summary field
				$data,
				$dataFileName,
				$writeToFile);

			// Copy all tags
			$sql = 'INSERT INTO "#message_revision_tag"
				("MessageId", "RevisionNumber", "TagId", "TagData")
				SELECT "MessageId", ? "RevisionNumber", "TagId", "TagData"
				FROM "#message_revision_tag"
				WHERE "MessageId" = ? AND "RevisionNumber" = ?';
			$parameters = array(
				$newRevision->GetRevisionNumber(),
				$this->messageId,
				$this->revisionNumber);
			$db->Execute($sql, $parameters);

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		return $newRevision;
	}

#	// Duplicates a message revision for later manipulation.
#	//
#	// The newly created message revision instance is also stored in the database. The moderation
#	// state of the original revision, tags and file data (in database / on disk) are copied.
#	//
#	// Throws {{UnbSecurityException}} If not logged in or no read/alter access on the message.
#	// Throws {{UnbDatabaseException}}
#	// Throws {{Exception}} If the attached file on disk could not be copied.
#	//
#	// Returns: The newly created UnbMessageRevision object, if successful, {{null}} otherwise.
#	//
#	// Privileges: Must be logged in and need read and alter access on the message
#	//
#	public function Duplicate()
#	{
#		if (!UnbSession::IsLoggedIn())
#			throw new UnbSecurityException('You must be logged in to create a message revision.');
#
#		$now = time();
#
#		// Check for read and alter access on the message
#		if (!UnbMessage::IsReadAllowedById($this->messageId))
#			throw new UnbSecurityException('You are not allowed to read this message.');
#		if (!UnbMessage::IsAlterAllowedById($this->messageId))
#			throw new UnbSecurityException('You are not allowed to alter this message.');
#
#		$db = UnbDatabase::GetInstance();
#		$db->StartTransaction(true);
#		try
#		{
#			$revisionNumber = $db->QueryValue(
#				'SELECT MAX("RevisionNumber") FROM "#message_revision" WHERE "MessageId" = ? FOR UPDATE',
#				$this->messageId);
#			if (!isset($revisionNumber))
#				$revisionNumber = 0;
#			$revisionNumber++;
#
#			$sql = 'INSERT INTO "#message_revision"
#				("MessageId", "RevisionNumber", "CreatedTime", "Author", "Subject", "Content", "Summary", "ModerationState", "Data", "HasData")
#				VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)';
#			$parameters = array(
#				$this->messageId,
#				$revisionNumber,
#				UnbDatabase::ConvertDateTime($now),
#				UnbSession::GetSessionUserId(),
#				$this->subject != '' ? $this->subject : null,
#				$this->content,
#				/*summary*/ null,
#				$this->moderationState,
#				$this->GetFileData(2),   // Only duplicate in-database data
#				$this->hasData);
#			$db->Execute($sql, $parameters);
#
#			// Copy all tags
#			$sql = 'INSERT INTO "#message_revision_tag"
#				("MessageId", "RevisionNumber", "TagId", "TagData")
#				SELECT "MessageId", ?, "TagId", "TagData"
#				FROM "#message_revision_tag"
#				WHERE "MessageId" = ? AND "RevisionNumber" = ?';
#			$parameters = array(
#				$revisionNumber,
#				$this->messageId,
#				$this->revisionNumber);
#			$db->Execute($sql, $parameters);
#
#			UnbMessage::UpdateSearchFields($messageId);
#
#			if ($this->hasData)
#			{
#				// Duplicate attached file on disk, if present
#				$filename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
#				if (file_exists($filename))
#				{
#					$newFilename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $revisionNumber;
#					if (!copy($filename, $newFilename))
#						throw new Exception('Could not duplicate attached file on disk.');
#				}
#			}
#
#			$db->Commit();
#		}
#		catch (Exception $ex)
#		{
#			$db->RollBack();
#			throw $ex;
#		}
#
#		$messageRevision = new self();
#		$messageRevision->messageId = $messageId;
#		$messageRevision->revisionNumber = $revisionNumber;
#		$messageRevision->createdTime = $now;
#		$messageRevision->author = UnbSession::GetSessionUserId();
#		$messageRevision->subject = $this->subject != '' ? $this->subject : null;
#		$messageRevision->content = $this->content;
#		$messageRevision->summary = null;
#		$messageRevision->moderationState = $this->moderationState;
#		$messageRevision->hasData = $this->hasData;
#		return $messageRevision;
#	}

	// Creates a UnbMessageRevision object from a database record array.
	//
	// rec = (array) Database record
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	//
	// Returns (UnbMessageRevision) object constructed from the database record
	//
	protected static function FromRecord($rec)
	{
		if (!is_array($rec))
			throw new InvalidArgumentException('Invalid argument type: $rec');
		if (!count($rec))
			throw new InvalidArgumentException('Record array is empty');

		$messageRevision = new self();
		$messageRevision->messageId = intval($rec['MessageId']);
		$messageRevision->revisionNumber = intval($rec['RevisionNumber']);
		$messageRevision->createdTime = UnbDatabase::ConvertDateTime($rec['CreatedTime']);
		$messageRevision->author = intval($rec['Author']);
		$messageRevision->subject = $rec['Subject'];
		$messageRevision->content = $rec['Content'];
		$messageRevision->htmlContent = $rec['HtmlContent'];
		$messageRevision->summary = $rec['Summary'];
		$messageRevision->moderationState = intval($rec['ModerationState']);
		$messageRevision->hasData = $rec['HasData'] && $rec['HasData'] !== 'FALSE';
		return $messageRevision;
	}

	// Reads a message revision from the database.
	//
	// This function is kind of a "service provider" for UnbMessageRevision::FindByMessageId()
	// and UnbMessageRevision::FindLatestByMessageId().
	//
	// messageId = (int/UnbMessage) Message or its ID to read
	// revisionNumber = (int) Revision number
	// contentLevel = (int) See UnbMessageRevision::FindByMessageId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	private static function ReadInternal($messageId, $revisionNumber, $contentLevel, $allowFallback)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');
		// NOTE: revisionNumber is checked in the function that calls this function
		if (!isset($contentLevel))
			$contentLevel = UnbMessageRevisionContentLevel::Content;
		if (!is_int($contentLevel))
			throw new InvalidArgumentException('Invalid argument type: $contentLevel');
		if ($contentLevel < 0)
			throw new InvalidArgumentException('Argument out of range: $contentLevel');

		$contentFields = '';
		if ($contentLevel == UnbMessageRevisionContentLevel::Subject)
			$contentFields .= 'mr."Subject", mr."Summary", ';
		elseif ($contentLevel == UnbMessageRevisionContentLevel::Content)
			$contentFields .= 'mr."Subject", mr."Summary", mr."Content", ';
		elseif ($contentLevel == UnbMessageRevisionContentLevel::HtmlContent)
			$contentFields .= 'mr."Subject", mr."Summary", mr."Content", mr."HtmlContent", ';

		$sessionUser = UnbSession::GetSessionUser();
		$sessionUserId = UnbSession::GetSessionUserId();

		if ($allowFallback)
		{
			// Probe all messages with the least content level that can be fallen back to.
			// But first, remember what content level we would like to get at best.
			$maxContentLevel = $contentLevel;
			$contentLevel = UnbMessageRevisionContentLevel::NoContent;
		}

		// Is reading the message (by the current session user) generally allowed?
		$isReadAllowed = UnbMessage::IsReadAllowedById($messageId);
		// If not:
		// Administrators and the message owner must be able to get the revision object to see the
		// tags and be able to change their access key lists. This is only allowed when no content
		// more than the subject (for handling convenience) is requested.  (Message owners are
		// handled directly in SQL, not here.)
		$isMsgAccessException = UnbSession::IsAdministrator() && $contentLevel <= UnbMessageRevisionContentLevel::Subject;

		// This is an exception value that determines the conditions when the current session user
		// will get access to a message revision when the message has restricting moderation flags
		// set. (Message owners are handled directly in SQL, not here.)
		$isMsgModException = UnbSession::IsModerator() ||
			$contentLevel == UnbMessageRevisionContentLevel::NoContent ||
			UnbSession::IsAdministrator() && $contentLevel <= UnbMessageRevisionContentLevel::Subject;
		// TODO: Administrators may also be allowed to read hidden/locked/unapproved message revisions.
		//       If so, add this:
		//       $isMsgModException |= UnbSession::IsAdministrator();

		$sql = 'SELECT mr."MessageId", mr."RevisionNumber", mr."CreatedTime", mr."Author", ' . $contentFields . 'mr."ModerationState", mr."HasData", m."ModerationFlags"
			FROM "#message" m
				JOIN "#message_revision" mr USING ("MessageId")
			WHERE mr."MessageId" = ? AND

				(? OR   -- Message access by keylists OR Exempted user condition for message access by keylists
				m."Owner" = ? AND ?) AND   -- Condition for message owner to get access to the message

				(m."ModerationFlags" & ' . (UnbMessageModerationFlag::Locked | UnbMessageModerationFlag::Hidden) . ' = 0 OR
				? OR   -- Exempted user condition for message access by moderation flags
				m."Owner" = ?) AND
				mr."RevisionNumber" = ';

		$parameters = array(
			$messageId,

			$isReadAllowed || $isMsgAccessException,   // Message access by keylists OR Exempted user condition for message access by keylists
			$sessionUserId,
			$contentLevel <= UnbMessageRevisionContentLevel::Subject,   // Condition for message owner to get access to the message

			$isMsgModException,   // Exempted user condition for message access by moderation flags
			$sessionUserId);

		// This is an exception value that determines the conditions when the current session user
		// will get access to a message revision when the message revision has a restricting
		// moderation state. (Revision authors are handled directly in SQL, not here.)
		#$isRevModException = UnbSession::IsModerator() ||
		#	$contentLevel == UnbMessageRevisionContentLevel::NoContent ||
		#	UnbSession::IsAdministrator() && $contentLevel <= UnbMessageRevisionContentLevel::Subject;
		$isRevModException = $isMsgModException;

		if (isset($revisionNumber))
		{
			// Continue the query like we know what revision number we're looking for
			$sql .= '? AND
					(mr."ModerationState" = ' . UnbMessageRevisionModerationState::Approved . ' OR
					? OR   -- This will be a boolean value indicating whether the sessionUser is an exempted user
					mr."Author" = ?)';
			$parameters[] = $revisionNumber;
			$parameters[] = $isRevModException;
			$parameters[] = $sessionUserId;
		}
		else
		{
			// Continue the query like get us the latest accessible revision
			$sql .= '(SELECT MAX(mr2."RevisionNumber")
						FROM "#message_revision" mr2
						WHERE mr2."MessageId" = mr."MessageId" AND
							(mr2."ModerationState" = ' . UnbMessageRevisionModerationState::Approved . ' OR
							? OR   -- This will be a boolean value indicating whether the sessionUser is an exempted user
							mr2."Author" = ?))';
			$parameters[] = $isRevModException;
			$parameters[] = $sessionUserId;
		}

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord($sql, $parameters);
		if (!$rec) return null;

		if ($allowFallback)
		{
			// Now is the time to see what maximum content level we can reach for this message.
			// The data has already been read for the maximum desired content level, now remove
			// the fields that are actually forbidden to read.

			// Start with the assumption that everything is allowed. This will be narrowed down
			// with the following checks.
			$possibleContentLevel = UnbMessageRevisionContentLevel::HtmlContent;

			if (!$isReadAllowed)
			{
				// Message read access by keylists
				// The message owner is handled just like an administrator in this case.
				if (UnbSession::IsAdministrator() || intval($rec['Owner']) == $sessionUserId)
				{
					$possibleContentLevel = min(UnbMessageRevisionContentLevel::Subject, $possibleContentLevel);
				}
				else
				{
					$possibleContentLevel = UnbMessageRevisionContentLevel::NoContent;
				}
			}

			if (intval($rec['ModerationFlags']) & (UnbMessageModerationFlag::Locked | UnbMessageModerationFlag::Hidden) ||
				intval($rec['ModerationState']) != UnbMessageRevisionModerationState::Approved)
			{
				// At least one of these: (they have the same effects)
				// * Message.ModerationFlags contains Hidden or Locked
				// * MessageRevision.ModerationState is not Approved (either Waiting or Locked)
				if (UnbSession::IsModerator() || intval($rec['Author']) == $sessionUserId)
				{
					// No restriction
				}
				elseif (UnbSession::IsAdministrator())
				{
					$possibleContentLevel = min(UnbMessageRevisionContentLevel::Subject, $possibleContentLevel);
				}
				else
				{
					$possibleContentLevel = UnbMessageRevisionContentLevel::NoContent;
				}
			}

			// Now apply the maximum possible AND desired content level and clear excess fields
			$contentLevel = min($possibleContentLevel, $maxContentLevel);
			if ($contentLevel < UnbMessageRevisionContentLevel::Subject)
				$rec['Subject'] = $rec['Summary'] = null;
			if ($contentLevel < UnbMessageRevisionContentLevel::Content)
				$rec['Content'] = null;
			if ($contentLevel < UnbMessageRevisionContentLevel::HtmlContent)
				$rec['HtmlContent'] = null;
		}

		$rev = self::FromRecord($rec);
		$rev->contentLevel = $contentLevel;
		return $rev;
	}

	// Finds a message revision instance by its message ID and revision number.
	//
	// messageId = (int/UnbMessage) Message or its ID of which to read the revision
	// revisionNumber = (int) Revision number
	// contentLevel = (int) Specifies whether the content should be retrieved from the database, too.
	//                     This can be one of the following values:
	//                     * UnbMessageRevisionContentLevel::NoContent - Don't retrieve any content
	//                     * UnbMessageRevisionContentLevel::Subject - Retrieve the subject and summary only
	//                     * UnbMessageRevisionContentLevel::Content - Also retrieve the normal content field (default if unset)
	//                     * UnbMessageRevisionContentLevel::HtmlContent - Retrieve the normal content field and its cached HTML version.
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FindByMessageId($messageId, $revisionNumber, $contentLevel = null, $allowFallback = false)
	{
		if (!is_int($revisionNumber))
			throw new InvalidArgumentException('Invalid argument type: $revisionNumber');
		if ($revisionNumber <= 0)
			throw new InvalidArgumentException('Argument out of range: $revisionNumber');

		return self::ReadInternal($messageId, $revisionNumber, $contentLevel, $allowFallback);
	}

	// Finds the latest accessible message revision instance by its message ID and moderation state.
	//
	// messageId = (int/UnbMessage) Message or its ID of which to read the revision
	// contentLevel = (int) See UnbMessageRevision::FindByMessageId()
	// allowFallback = (bool) Allow falling back to a lower content level if moderation restrictions require it
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: A UnbMessageRevision object, if successful, {{null}} otherwise.
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FindLatestByMessageId($messageId, $contentLevel = null, $allowFallback = false)
	{
		return self::ReadInternal($messageId, null, $contentLevel, $allowFallback);
	}

	// Gets the time when the message's first revision was created.
	//
	// TODO: Deprecated because of UnbMessage::GetCreatedTime() ?
	//
	// messageId = (int/UnbMessage) Message or its ID of which to get the creation time
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbDatabaseException}}
	//
	// Returns (int) UNIX timestamp
	//
	// Privileges: Need read access on the message. Additionally, the message must not have a
	// moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FirstCreatedTime($messageId)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		// Is reading the message (by the current session user) generally allowed?
		$isReadAllowed = UnbMessage::IsReadAllowedById($messageId);
		// If not:
		// Administrators and the message owner must be able to get the revision object to see the
		// tags and be able to change their access key lists. This is only allowed when no content
		// more than the subject (for handling convenience) is requested. (Message owners are
		// handled directly in SQL, not here.)
		$isMsgAccessException = UnbSession::IsAdministrator();

		// This is an exception value that determines the conditions when the current session user
		// will get access to the message when the message has restricting moderation flags set.
		// (Message owners are handled directly in SQL, not here.)
		$isMsgModException = UnbSession::IsModerator() ||
			UnbSession::IsAdministrator();

		$sql = 'SELECT mr."CreatedTime", m."Owner"
			FROM "#message" m
				JOIN "#message_revision" mr USING ("MessageId")
			WHERE mr."MessageId" = ? AND

				(? OR   -- Message access by keylists OR Exempted user condition for message access by keylists
				m."Owner" = ?) AND

				(m."ModerationFlags" & ' . UnbMessageModerationFlag::Hidden . ' = 0 OR
				? OR   -- This will be a boolean value indicating whether the sessionUser is an exempted user
				m."Owner" = ?) AND
				mr."RevisionNumber" =
					(SELECT MIN(mr2."RevisionNumber")
					FROM "#message_revision" mr2
					WHERE mr2."MessageId" = mr."MessageId")';
		$parameters = array(
			$messageId,

			$isReadAllowed || $isMsgAccessException,   // Message access by keylists OR Exempted user condition for message access by keylists
			UnbSession::GetSessionUserId(),

			$isMsgModException,
			UnbSession::GetSessionUserId());

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord($sql, $parameters);
		if (!isset($rec)) return null;
		return UnbDatabase::ConvertDateTime($rec['CreatedTime']);
	}

	// Finds all message revisions by their message ID.
	//
	// messageId = (int/UnbMessage) Message or its ID of which to read the revision
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: Array of UnbMessageRevision objects, if successful, {{null}} otherwise.
	//
	// Privileges: Need read access on the message. Additionally, message and message revision
	// must not have a moderation lock of any kind. Exceptions are documented separately.
	//
	public static function FindAllByMessageId($messageId)
	{
		if ($messageId instanceof UnbMessage)
			$messageId = $messageId->GetMessageId();
		if (!is_int($messageId))
			throw new InvalidArgumentException('Invalid argument type: $messageId');
		if ($messageId <= 0)
			throw new InvalidArgumentException('Argument out of range: $messageId');

		// Is reading the message (by the current session user) generally allowed?
		$isReadAllowed = UnbMessage::IsReadAllowedById($messageId);
		// If not:
		// Administrators and the message owner must be able to get the revision object to see the
		// tags and be able to change their access key lists. This is only allowed when no content
		// more than the subject (for handling convenience) is requested. (Message owners are
		// handled directly in SQL, not here.)
		$isMsgAccessException = UnbSession::IsAdministrator();

		$sql = 'SELECT mr."MessageId", mr."RevisionNumber", mr."CreatedTime", mr."Author", mr."Summary", mr."ModerationState"
			FROM "#message" m
				JOIN "#message_revision" mr USING ("MessageId")
			WHERE mr."MessageId" = ? AND

				(? OR   -- Message access by keylists OR Exempted user condition for message access by keylists
				m."Owner" = ?)

			ORDER BY mr."RevisionNumber"';

		$parameters = array(
			$messageId,

			$isReadAllowed || $isMsgAccessException,   // Message access by keylists OR Exempted user condition for message access by keylists
			UnbSession::GetSessionUserId());

		$db = UnbDatabase::GetInstance();
		$recs = $db->Query($sql, $parameters);
		if (!isset($recs)) return null;
		$revs = array();
		foreach ($recs as $rec)
		{
			$contentLevel = UnbMessageRevisionContentLevel::Subject;   // NOTE: This also includes the subject, but it's unset here.

			if ($rec['ModerationState'] != UnbMessageRevisionModerationState::Approved &&
				!UnbSession::IsModerator() &&
				UnbSession::GetSessionUserId() != intval($rec['Author']))
			{
				// This revision is not approved and we're neither a moderator nor the revision's author.
				// Hide all user-content fields.
				$rec['Summary'] = null;
				$contentLevel = UnbMessageRevisionContentLevel::NoContent;
			}

			$rev = self::FromRecord($rec);
			$rev->contentLevel = $contentLevel;
			$revs[] = $rev;
		}
		return $revs;
	}

	// Gets the revision's content level.
	//
	// Returns (int) Content level value, see UnbMessageRevisionContentLevel constants.
	//
	public function GetContentLevel()
	{
		return $this->contentLevel;
	}

	// Gets the revision's message ID.
	//
	// Returns (int) Message ID
	//
	public function GetMessageId()
	{
		return $this->messageId;
	}

	// Gets the revision number.
	//
	// Returns (int) Revision number
	//
	public function GetRevisionNumber()
	{
		return $this->revisionNumber;
	}

	// Gets the time when the revision was created.
	//
	// Returns (int) UNIX timestamp
	//
	public function GetCreatedTime()
	{
		return $this->createdTime;
	}

	// Gets the user ID of the author who created this revision.
	//
	// Returns (int) Author user ID
	//
	public function GetAuthorUserId()
	{
		return $this->author;
	}

	// Gets the UnbUser object of the message revision's author.
	//
	// Returns: (UnbUser) Author user
	//
	public function GetAuthorUser()
	{
		if (isset($this->authorUser))
			return $this->authorUser;
		$this->authorUser = UnbUser::FindByUserId($this->author);
		return $this->authorUser;
	}

#	If enabled, this method needs a documentation with parameter, exceptions and return value
#	description and argument type checking. (And exception handling and privileges)
#
#	public function SetAuthor($author)
#	{
#		$this->db->StartTransaction();
#		$r = $this->db->Execute(
#			'UPDATE "#message_revision" SET "Author" = ? WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#			array($author, $this->messageId, $this->revisionNumber));
#		if (!$r)
#		{
#			$this->db->RollBack();
#			return false;
#		}
#		$r = $this->db->Commit();
#		if ($r)
#		{
#			$this->author = $author;
#			$this->authorUser = null;   // TODO: Save if passed as argument
#		}
#		return $r;
#	}

	// Gets the message revision's subject.
	//
	// Returns (string) Subject
	//
	public function GetSubject()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Subject)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the subject.');

		return $this->subject;
	}

#	// Sets the message revision's subject.
#	//
#	// subject = (string) Subject of the message's first revision or {{null}} (max. 255 characters)
#	//
#	// Throws {{InvalidArgumentException}} On invalid argument type or value.
#	// Throws {{LengthException}}
#	// Throws {{UnbSecurityException}} If not logged in or not the same user/no moderator.
#	// Throws {{UnbDatabaseException}}
#	//
#	// Privileges: Must be logged in as the same user or with moderator privilege.
#	//
#	public function SetSubject($subject)
#	{
#		if (isset($subject))
#		{
#			if (!is_string($subject))
#				throw new InvalidArgumentException('Invalid argument type: $subject');
#			$subject = trim($subject);
#			if (mb_strlen($subject) > 255)
#				throw new LengthException('[#4] Subject is too long');
#			if (!strlen($subject)) $subject = null;
#		}
#
#		if (!UnbSession::IsModerator())
#			throw new UnbSecurityException('You need the moderator privilege to set the subject.');
#
#		$this->db->StartTransaction(true);   // lock required for UpdateSearchFields
#		try
#		{
#			$this->db->Execute(
#				'UPDATE "#message_revision" SET "Subject" = ? WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#				array($subject, $this->messageId, $this->revisionNumber));
#
#			UnbMessage::UpdateSearchFields($this->messageId);
#
#			$this->db->Commit();
#		}
#		catch (Exception $ex)
#		{
#			$db->RollBack();
#			throw $ex;
#		}
#		$this->subject = $subject;
#	}

	// Gets the message revision's full content (if retrieved).
	//
	// Returns (string) Content
	//
	public function GetContent()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the content.');

		return $this->content;
	}

#	// Sets the message revision's content.
#	//
#	// content = (string) Content of the message's first revision (max. 16,777,215 characters)
#	//
#	// Throws {{InvalidArgumentException}} On invalid argument type or value.
#	// Throws {{LengthException}}
#	// Throws {{UnbSecurityException}} If not logged in or not the same user/no moderator.
#	// Throws {{UnbDatabaseException}}
#	//
#	// Privileges: Must be logged in as the same user or with moderator privilege.
#	//
#	public function SetContent($content)
#	{
#		if (isset($content))
#		{
#			if (!is_string($content))
#				throw new InvalidArgumentException('Invalid argument type: $content');
#			$content = trim($content);
#			if (mb_strlen($content) > 16777215)
#				throw new LengthException('[#3] Content is too long');
#			if (!strlen($content)) $content = null;
#		}
#
#		if (!UnbSession::IsModerator())
#			throw new UnbSecurityException('You need the moderator privilege to set the content.');
#
#		$this->db->StartTransaction(true);   // lock required for UpdateSearchFields
#		try
#		{
#			$this->db->Execute(
#				'UPDATE "#message_revision" SET "Content" = ?, "HtmlContent" = NULL WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#				array($content, $this->messageId, $this->revisionNumber));
#
#			UnbMessage::UpdateSearchFields($this->messageId);
#
#			$this->db->Commit();
#		}
#		catch (Exception $ex)
#		{
#			$db->RollBack();
#			throw $ex;
#		}
#		$this->content = $content;
#	}

	// Gets the message revision's cached HTML version of the content (if available and retrieved).
	//
	// autoGenerate = (bool) Automatically generate and store the HTML content if it is not available yet
	// markup = (UnbMarkup) Markup processor to use for generating HTML content
	//
	// Returns (string) HTML content
	//
	public function GetHtmlContent($autoGenerate = false, $markup = null)
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::HtmlContent)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the HTML content.');

		if (!isset($this->htmlContent) && $autoGenerate)
		{
			if ($this->GetContent() !== null)
			{
				$html = $markup->Convert($this->GetContent());
			}
			else
			{
				$html = '';   // Empty contents should always be empty HTML
			}
			try
			{
				$this->SetHtmlContent($html);
			}
			catch (Exception $ex)
			{
				// Ignore errors here
			}
		}

		return $this->htmlContent;
	}

	// Sets the message revision's cached HTML version of the content.
	//
	// htmlContent = (string) HTML content or {{null}} (max. 16,777,215 characters)
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{UnbDatabaseException}}
	//
	public function SetHtmlContent($htmlContent)
	{
		if (isset($htmlContent))
		{
			if (!is_string($htmlContent))
				throw new InvalidArgumentException('Invalid argument type: $htmlContent');
			$htmlContent = trim($htmlContent);
			if (mb_strlen($htmlContent) > 16777215)
				throw new LengthException('[#3] HTML content is too long');
			if (!strlen($htmlContent))
				$htmlContent = null;
		}

		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message_revision" SET "HtmlContent" = ? WHERE "MessageId" = ? AND "RevisionNumber" = ?',
				array($htmlContent, $this->messageId, $this->revisionNumber));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		$this->htmlContent = $htmlContent;
	}

	// Gets this revision's summary of changes to the previous one.
	//
	// Returns (string) Summary
	//
	public function GetSummary()
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Subject)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the summary.');

		return $this->summary;
	}

#	// Sets this revision's summary of changes to the previous one.
#	//
#	// summary = (string) Summary of the changes in the new revivion to the previous one or {{null}} (max. 255 characters)
#	//
#	// Throws {{InvalidArgumentException}} On invalid argument type or value.
#	// Throws {{LengthException}}
#	// Throws {{UnbSecurityException}} If not logged in or no moderator.
#	// Throws {{UnbDatabaseException}}
#	//
#	// Privileges: Must be logged in with moderator privilege
#	//
#	public function SetSummary($summary)
#	{
#		if (isset($summary))
#		{
#			if (!is_string($summary))
#				throw new InvalidArgumentException('Invalid argument type: $summary');
#			$summary = trim($summary);
#			if (mb_strlen($summary) > 255)
#				throw new LengthException('[#5] Summary is too long');
#			if (!strlen($summary)) $summary = null;
#		}
#
#		if (!UnbSession::IsModerator())
#			throw new UnbSecurityException('You need the moderator privilege to set the summary.');
#
#		$this->db->StartTransaction();
#		try
#		{
#			$this->db->Execute(
#				'UPDATE "#message_revision" SET "Summary" = ? WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#				array($summary, $this->messageId, $this->revisionNumber));
#			$this->db->Commit();
#		}
#		catch (Exception $ex)
#		{
#			$db->RollBack();
#			throw $ex;
#		}
#		$this->summary = $summary;
#	}

	// Gets a value indicating whether this message revision has data attached.
	//
	// Returns: (bool) Message revision has data attached
	//
	public function HasFileData()
	{
		return $this->hasData;
	}

	// Gets the length of an attachment file in bytes.
	//
	// searchMode = (int) Specifies how to read the data.
	//                    * 0: Find the data whereever it is, look in the database first.
	//                    * 1: Find the data whereever it is, look in the filesystem first.
	//
	// Throws {{UnbDatabaseException}}
	// Throws {{Exception}} If the attached data was not found although it's supposed to be there.
	//
	// Returns the size of the attached file in bytes, or 0 if there is no file attached.
	//
	public function GetFileDataLength($searchMode = 0)
	{
		if (!$this->hasData) return 0;

		if ($searchMode == 0)
		{
			$val = $this->db->QueryValue(
				'SELECT LENGTH("Data") FROM "#message_revision"
				WHERE "MessageId" = ? AND "RevisionNumber" = ?',
				array($this->messageId, $this->revisionNumber));
		}
		if (!isset($val))
		{
			// There is no data in the database, look for the local file
			$filename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
			if (file_exists($filename))
			{
				$val = filesize($filename);
			}
			else
			{
				if ($searchMode == 1)
				{
					$val = $this->db->QueryValue(
						'SELECT LENGTH("Data") FROM "#message_revision"
						WHERE "MessageId" = ? AND "RevisionNumber" = ?',
						array($this->messageId, $this->revisionNumber));
				}
				if (!isset($val))
					throw new Exception('[#9] Attached data not found.');
			}
		}
		return $val;
	}

	// Gets the name of the attachment file on disk.
	//
	// Throws {{UnbDatabaseException}}
	//
	// This method only returns a file name when the file exists. It is not verified that the
	// attached data exists in the database otherwise. If the constructed file name does not
	// exist, it will not be returned.
	//
	// Returns the name of the attached file, or {{null}} if there is no file attached.
	//
	public function GetFileDataPath()
	{
		if (!$this->hasData) return null;

		$filename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
		if (file_exists($filename))
		{
			return $filename;
		}
		return null;
	}

	// Gets the contents of an attachment file.
	//
	// searchMode = (int) Specifies how to read the data.
	//                    * 0: Find the data whereever it is, look in the database first.
	//                    * 1: Find the data whereever it is, look in the filesystem first.
	//                    * 2: Only read from the database, return {{null}} if it's not there.
	//                    * 3: Use {{readfile}} to read from the filesystem. Returns {{true}} or {{false}} depending on success.
	//
	// Throws {{UnbDatabaseException}}
	// Throws {{UnbSecurityException}} If the message revision was requested with NoContent mode.
	// Throws {{Exception}} If the attached data was not found although it's supposed to be there.
	//
	// Returns the contents of the attached file, or {{null}} if there is no file attached.
	// If searchMode is 3, the return value is a boolean value.
	//
	public function GetFileData($searchMode = 0)
	{
		if ($this->contentLevel < UnbMessageRevisionContentLevel::Content)
			throw new UnbSecurityException('This revision was not retrieved with an appropriate content level to get the file data.');

		if (!$this->hasData)
		{
			if ($searchMode == 3) return false;
			return null;
		}

		if ($searchMode == 0 || $searchMode == 2)
		{
			$val = $this->db->QueryValue(
				'SELECT "Data" FROM "#message_revision"
				WHERE "MessageId" = ? AND "RevisionNumber" = ?',
				array($this->messageId, $this->revisionNumber));
		}
		if (!isset($val))
		{
			if ($searchMode == 2) return null;

			// There is no data in the database, look for the local file
			$filename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
			if (file_exists($filename))
			{
				if ($searchMode == 3)
					return readfile($filename) !== false;

				$val = file_get_contents($filename);
			}
			else
			{
				if ($searchMode == 3)
					return false;   // readfile() doesn't work. Possible data in the database is irrelevant in this case.
				if ($searchMode == 1)
				{
					$val = $this->db->QueryValue(
						'SELECT "Data" FROM "#message_revision"
						WHERE "MessageId" = ? AND "RevisionNumber" = ?',
						array($this->messageId, $this->revisionNumber));
				}
				if (!isset($val))
					throw new Exception('[#9] Attached data not found.');
			}
		}
		return $val;
	}

#	// Stores the contents of an attachment file.
#	//
#	// data = (string) Attachment file contents to store (can be {{null}} if tmpFilename is used or to clear data)
#	// tmpFilename = (string) Temp filename from uploads, e.g.: $_FILES['userfile']['tmp_name']
#	// writeToFile = (bool) {{true}} to write the contents into a file instead of the database
#	//
#	// Throws {{InvalidArgumentException}} If the input file does not exist.
#	// Throws {{Exception}} If the input file could not be read or copied.
#	// Throws {{UnbSecurityException}} If not logged in or not the same user/no moderator.
#	// Throws {{UnbDatabaseException}}
#	//
#	// Privileges: Must be logged in as the same user or with moderator privilege.
#	//
#	public function SetFileData($data, $tmpFilename = null, $writeToFile = null)
#	{
#		if (isset($data))
#		{
#			if (!is_string($data))
#				throw new InvalidArgumentException('Invalid argument type: $data');
#		}
#		if (isset($tmpFilename))
#		{
#			if (!is_string($tmpFilename))
#				throw new InvalidArgumentException('Invalid argument type: $tmpFilename');
#			if (!strlen($tmpFilename)) $tmpFilename = null;
#			if (isset($tmpFilename) && !file_exists($tmpFilename))
#				throw new InvalidArgumentException('The specified file does not exist.');
#		}
#
#		if (UnbSession::GetSessionUserId() != $this->author && !UnbSession::IsModerator())
#			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to set file data.');
#
#		// We save new data if either field is set. If both are unset, data is cleared.
#		$newData = isset($data) || isset($tmpFilename);
#
#		$this->db->StartTransaction();
#		try
#		{
#			if ($newData)
#			{
#				if (!isset($writeToFile))
#				{
#					// No indication whether to save data in a file or in the database, ask our configuration
#					$dbMaxSize = UnbConfiguration::Get('attachment.storage.database-max-size', -1);
#					if ($dbMaxSize == -1)
#					{
#						// Always write to the database (default if unconfigured)
#						$writeToFile = false;
#					}
#					else
#					{
#						// Write to the database up to a limit, larger data goes into a file
#						if (is_string($data))
#							$dataLength = strlen($data);
#						else
#							$dataLength = filesize($tmpFilename);
#						$writeToFile = $dataLength > $dbMaxSize;
#					}
#				}
#
#				if ($writeToFile)
#				{
#					// Output is a file
#
#					// Clear present data in the database first
#					$this->db->Execute(
#						'UPDATE "#message_revision" SET "HasData" = ?, "Data" = NULL WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#						array(1, $this->messageId, $this->revisionNumber));
#
#					// Now write the new file
#					$newFilename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
#					if (is_string($tmpFilename))
#					{
#						// Input is a file
#						if (is_uploaded_file($tmpFilename))
#						{
#							if (!move_uploaded_file($tmpFilename, $newFilename))
#								throw new Exception('[#6] The uploaded file could not be copied to its new location.');
#						}
#						else
#						{
#							if (!copy($tmpFilename, $newFilename))
#								throw new Exception('[#6] The input file could not be copied to its new location.');
#						}
#					}
#					else
#					{
#						// Input is a string
#						if (file_put_contents($newFilename, $data) === false)
#							throw new Exception('[#7] The data could not be written to the local file.');
#					}
#				}
#				else
#				{
#					// Output is database
#					if (is_string($tmpFilename))
#					{
#						// Input is a file
#						$data = file_get_contents($tmpFilename);
#						if ($data === false)
#							throw new Exception('[#8] The data could not be read from the input file.');
#					}
#
#					$this->db->Execute(
#						'UPDATE "#message_revision" SET "HasData" = ?, "Data" = ? WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#						array(1, $data, $this->messageId, $this->revisionNumber));
#
#					// Delete present file if there is one
#					$oldFilename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
#					if (file_exists($oldFilename))
#						if (!unlink($oldFilename))
#							throw new Exception('[#10] The old file could not be deleted.');
#				}
#			}
#			else   // !$newData
#			{
#				// Any stored data shall be cleared
#				$this->db->Execute(
#					'UPDATE "#message_revision" SET "HasData" = ?, "Data" = NULL WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#					array(0, $this->messageId, $this->revisionNumber));
#
#				$oldFilename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
#				if (file_exists($oldFilename))
#					if (!unlink($oldFilename))
#						throw new Exception('[#10] The old file could not be deleted.');
#			}
#			$this->db->Commit();
#
#			$this->hasData = $newData;
#		}
#		catch (Exception $ex)
#		{
#			$this->db->RollBack();
#			throw $ex;
#		}
#	}

#	// Copies attached file data from the previous revision of this message.
#	//
#	// Throws {{Exception}} If the previous message revision could not be fetched. (If this is already the first revision, no exception will be thrown.)
#	// Throws {{UnbSecurityException}} If not logged in or not the same user/no moderator.
#	// Throws {{UnbDatabaseException}}
#	//
#	// Privileges: Must be logged in as the same user or with moderator privilege.
#	//
#	public function DuplicateFileData()
#	{
#		if ($this->revisionNumber == 1) return;   // There can't be a revision before the first
#
#		if (UnbSession::GetSessionUserId() != $this->author && !UnbSession::IsModerator())
#			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to duplicate file data.');
#
#		$prevRevNum = $this->revisionNumber - 1;
#
#		// Get information about previous message revision
#		$prevRev = self::FindByMessageId($this->messageId, $prevRevNum, UnbMessageRevisionContentLevel::Content);
#		if (!isset($prevRev))
#			throw new Exception('Cannot fetch previous message revision.');
#
#		$this->db->StartTransaction();
#		try
#		{
#			#$this->db->Execute(
#			#	'UPDATE "#message_revision"
#			#	SET "HasData" = ?, "Data" =
#			#		(SELECT "Data" FROM "#message_revision" WHERE "MessageId" = ? AND "RevisionNumber" = ?)
#			#	WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#			#	array(
#			#		$prevRev->HasFileData() ? 1 : 0,
#			#		$this->messageId,
#			#		$prevRevNum,
#			#		$this->messageId,
#			#		$this->revisionNumber));
#			// NOTE: MySQL doesn't currently support SELECTing from a table being UPDATEd. Sad but true.
#			// See http://dev.mysql.com/doc/refman/5.1/en/update.html (at the very bottom)
#
#			$this->db->Execute(
#				'UPDATE "#message_revision"
#				SET "HasData" = ?, "Data" = ?
#				WHERE "MessageId" = ? AND "RevisionNumber" = ?',
#				array(
#					$prevRev->HasFileData() ? 1 : 0,
#					$prevRev->GetFileData(2),
#					$this->messageId,
#					$this->revisionNumber));
#
#			// Duplicate attached file on disk, if present
#			$filename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $prevRevNum;
#			$newFilename = UnbConfiguration::Get('path.file-uploads', '.') . '/' . $this->messageId . '_' . $this->revisionNumber;
#			if (file_exists($filename))
#			{
#				if (!copy($filename, $newFilename))
#					throw new Exception('Could not duplicate attached file on disk.');
#			}
#			else
#			{
#				if (file_exists($newFilename))
#					if (!unlink($newFilename))
#						throw new Exception('[#10] Could not delete present attached file on disk.');
#			}
#
#			$this->db->Commit();
#
#			$this->hasData = $prevRev->HasFileData();
#		}
#		catch (Exception $ex)
#		{
#			$this->db->RollBack();
#			throw $ex;
#		}
#	}

	// Gets this revision's moderation state.
	//
	// Returns (int) Moderation state
	//
	public function GetModerationState()
	{
		return $this->moderationState;
	}

	// Sets this revision's moderation state.
	//
	// moderationState = (int) See the UnbMessageRevision::MS_* constants.
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function SetModerationState($moderationState)
	{
		if (!is_int($moderationState))
			throw new InvalidArgumentException('Invalid argument type: $moderationState');
		if ($moderationState < 0)
			throw new InvalidArgumentException('Argument out of range: $moderationState');

		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to set the moderation state.');

		$this->db->StartTransaction(true);
		try
		{
			$this->db->Execute(
				'UPDATE "#message_revision" SET "ModerationState" = ? WHERE "MessageId" = ? AND "RevisionNumber" = ?',
				array($moderationState, $this->messageId, $this->revisionNumber));

			UnbMessage::UpdateSearchFields($this->messageId);

			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		$this->moderationState = $moderationState;
	}

	// Deletes this instance of the message revision with all its references.
	//
	// Throws {{UnbSecurityException}} If not logged in or no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in with moderator privilege
	//
	public function Delete()
	{
		if (!UnbSession::IsModerator())
			throw new UnbSecurityException('You need the moderator privilege to delete a message revision.');

		$this->db->StartTransaction();
		try
		{
			// All references are automatically resolved by the database foreign key contraints. :)

			$this->db->Execute(
				'DELETE FROM "#message_revision" WHERE "MessageId" = ? AND "RevisionNumber" = ?',
				array($this->messageId, $this->revisionNumber));

			// Determine whether a file is attached to this message revision that needs to be deleted
			$dataFilename = $this->GetFileDataPath();
			if (file_exists($dataFilename))
				if (!unlink($dataFilename))
					throw new Exception('[#10] The attached file could not be deleted.');

			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Gets UnbTag objects of all tags assigned to this message revision.
	//
	// onlyVisible = (bool) Only get tags that don't have the UnbTagFlag::Invisible flag set
	// names = (array) Tag names to retrieve
	//
	// Throws {{UnbDatabaseException}}
	//
	// If a list of tag names was specified, the return value is an associative array with the
	// tags' display name as key. If no names were specified, all tags are returned in a normal
	// array indexed numerically sorted in the correct hierarchical tag order.
	//
	// Returns: (array(UnbTag)) Tags
	//
	public function GetTags($onlyVisible = true, $names = null)
	{
		if (isset($names))
		{
			if (!is_array($names))
				throw new InvalidArgumentException('Invalid argument type: $names');
		}

		// t.TagId would be ambiguous (from both tables), so an alias for this one is required
		$sql = 'SELECT t."TagId" AS "TagId", t.*, mrt."TagData"
			FROM "#message_revision_tag" mrt
				JOIN "#tag" t USING ("TagId")
			WHERE mrt."MessageId" = ? AND mrt."RevisionNumber" = ?';
		$parameters = array($this->messageId, $this->revisionNumber);

		if ($onlyVisible)
			$sql .= ' AND NOT (t.Flags & ' . UnbTagFlag::Invisible . ')';

		if (isset($names))
		{
			$sql .= ' AND (';
			$count = 0;
			foreach ($names as $name)
			{
				if (!is_string($name))
					throw new InvalidArgumentException('Invalid argument type: $names');

				if ($count++)
					$sql .= ' OR ';
				$sql .= 't.TagName = ?';
				$parameters[] = $name;
			}
			$sql .= ')';
		}

		#$sql .= ' ORDER BY t."TagName"';
		$order = UnbTag::GetHierarchicalOrder();
		if (count($order))
		{
			$sql .= ' ORDER BY CASE t."TagId"';
			foreach ($order as $index => $tagId)
				$sql .= ' WHEN ' . $tagId . ' THEN ' . $index;
			$sql .= ' END';
		}

		$recs = $this->db->Query($sql, $parameters);
		$tags = array();
		foreach ($recs as $rec)
		{
			if (isset($names))
			{
				$tag = UnbTag::FromRecord($rec);
				$tags[$tag->GetTagName()] = $tag;
			}
			else
			{
				$tags[] = UnbTag::FromRecord($rec);
			}
		}
		return $tags;
	}

	// Gets (unsorted) tag IDs of all tags assigned to this message revision.
	//
	// withData = (bool) {{true}} to include tag data, {{false}} otherwise.
	//
	// Throws {{UnbDatabaseException}}
	//
	// Returns: An array of all tag ID numbers. If {{withData}} is set, a tag has the UserData
	// flag set and tag data stored (i.e. not null), the array item will be an
	// array(TagID => (int), TagData => (string)) instead of the (int) number only.
	//
	public function GetTagIds($withData = true)
	{
		$recs = $this->db->Query(
			'SELECT "TagId", "TagData" FROM "#message_revision_tag" WHERE "MessageId" = ? AND "RevisionNumber" = ?',
			array($this->messageId, $this->revisionNumber));
		$tags = array();
		foreach ($recs as $rec)
		{
			$rec['TagId'] = intval($rec['TagId']);
			if ($withData && $rec['TagData'] != null)
			{
				$tags[] = $rec;
			}
			else
			{
				$tags[] = $rec['TagId'];
			}
		}
		return $tags;
	}

	// Assigns a tag to this message revision.
	//
	// tagId = (int/UnbTag) Tag or its ID to assign.
	// tagData = (string) Tag data to use for this message revision. (Only for tags with UserData flag,
	//                    {{null}} otherwise. Max. 255 characters)
	//
	// Throws {{InvalidArgumentException}}
	// Throws {{LengthException}}
	// Throws {{UnbSecurityException}} If not logged in or not the same user/no moderator or tag use disallowed.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the same user or with moderator privilege. You must be allowed
	// to use the specified tag.
	//
	// TODO: Reset moderation state when a UserData tag has been altered - when? (Except for moderators)
	//
	public function AddTag($tagId, $tagData = null)
	{
		if ($tagId instanceof UnbTag)
			$tagId = $tagId->GetTagId();
		if (!is_int($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');
		if ($tagId <= 0)
			throw new InvalidArgumentException('Argument out of range: $tagId');
		if (isset($tagData))
		{
			if (!is_string($tagData))
				throw new InvalidArgumentException('Invalid argument type: $tagData');
			$tagData = trim($tagData);
			if (mb_strlen($tagData) > 255)
				throw new LengthException('[#11] Tag data is too long');
		}

		if (UnbSession::GetSessionUserId() != $this->author && !UnbSession::IsModerator())
			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to add a tag.');
		if (!UnbTag::IsUseAllowedById($tagId))
			throw new UnbSecurityException('You are not allowed to use this tag.');

		UnbMessage::ClearAccessCache();
		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'UPDATE "#message_revision_tag"
				SET "TagData" = ?
				WHERE "MessageId" = ? AND "RevisionNumber" = ? AND "TagId" = ?',
				array($tagData, $this->messageId, $this->revisionNumber, $tagId));
			if ($this->db->GetRowsAffected() == 0)
			{
				// The entry doesn't exist yet, so insert it.
				try
				{
					$this->db->Execute(
						'INSERT INTO "#message_revision_tag" ("MessageId", "RevisionNumber", "TagId", "TagData") VALUES (?, ?, ?, ?)',
						array($this->messageId, $this->revisionNumber, $tagId, $tagData));
				}
				catch (UnbDatabaseException $ex)
				{
					// If the entry already exists, it is unchanged.
					// GetRowsAffected() has returned 0 in this case although the record has been found.
					// But it has not been altered because it is unchanged.
					if (!$this->db->IsConstraintError()) throw $ex;
				}
			}
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

	// Removes a tag from this message revision.
	//
	// tagId = (int/UnbTag) Tag or its ID to remove.
	//
	// Throws {{InvalidArgumentException}} On invalid argument type or value.
	// Throws {{UnbSecurityException}} If not logged in or not the same user/no moderator.
	// Throws {{UnbDatabaseException}}
	//
	// Privileges: Must be logged in as the same user or with moderator privilege
	//
	public function RemoveTag($tagId)
	{
		if ($tagId instanceof UnbTag)
			$tagId = $tagId->GetTagId();
		if (!is_int($tagId))
			throw new InvalidArgumentException('Invalid argument type: $tagId');
		if ($tagId <= 0)
			throw new InvalidArgumentException('Argument out of range: $tagId');

		if (UnbSession::GetSessionUserId() != $this->author && !UnbSession::IsModerator())
			throw new UnbSecurityException('You must be logged in as this user or need the moderator privilege to remove a tag.');

		UnbMessage::ClearAccessCache();
		$this->db->StartTransaction();
		try
		{
			$this->db->Execute(
				'DELETE FROM "#message_revision_tag" WHERE "MessageId" = ? AND "RevisionNumber" = ? AND "TagId" = ?',
				array($this->messageId, $this->revisionNumber, $tagId));
			$this->db->Commit();
		}
		catch (Exception $ex)
		{
			$this->db->RollBack();
			throw $ex;
		}
	}

} // class UnbMessageRevision

?>